import filecmp

from .. import api

description = """
Compare the contents of two files.

This command will compare the contents of two files, returning 'True' if they
are identical and 'False' otherwise.
"""

epilog = f"""
[Example] Compare two files:
  $ fuc {api.common._script_name()} left.txt right.txt
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Compare the contents of two files.',
    )
    parser.add_argument('left', help='Left file.')
    parser.add_argument('right', help='Right file.')

def main(args):
    print(filecmp.cmp(args.left, args.right))



import os
import sys
import shutil

from .. import api

import pandas as pd

description = f"""
Pipeline for converting FASTQ files to analysis-ready BAM files.

Here, "analysis-ready" means that the final BAM files are: 1) aligned to a
reference genome, 2) sorted by genomic coordinate, 3) marked for duplicate
reads, 4) recalibrated by BQSR model, and 5) ready for downstream analyses
such as variant calling.

External dependencies:
  - SGE: Required for job submission (i.e. qsub).
  - BWA: Required for read alignment (i.e. BWA-MEM).
  - SAMtools: Required for sorting and indexing BAM files.
  - GATK: Required for marking duplicate reads and recalibrating BAM files.

Manifest columns:
  - Name: Sample name.
  - Read1: Path to forward FASTA file.
  - Read2: Path to reverse FASTA file.
"""

epilog = f"""
[Example] Specify queue:
  $ fuc {api.common._script_name()} \\
  manifest.csv \\
  ref.fa \\
  output_dir \\
  "-q queue_name -pe pe_name 10" \\
  "-q queue_name" \\
  "-Xmx15g -Xms15g" \\
  1.vcf 2.vcf 3.vcf \\
  --thread 10

[Example] Specify nodes:
  $ fuc {api.common._script_name()} \\
  manifest.csv \\
  ref.fa \\
  output_dir \\
  "-l h='node_A|node_B' -pe pe_name 10" \\
  "-l h='node_A|node_B'" \\
  "-Xmx15g -Xms15g" \\
  1.vcf 2.vcf 3.vcf \\
  --thread 10
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Pipeline for converting FASTQ files to analysis-ready BAM files.',
    )
    parser.add_argument(
        'manifest',
        help='Sample manifest CSV file.'
    )
    parser.add_argument(
        'fasta',
        help='Reference FASTA file.'
    )
    parser.add_argument(
        'output',
        type=os.path.abspath,
        help='Output directory.'
    )
    parser.add_argument(
        'qsub1',
        type=str,
        help="SGE resoruce to request with qsub for read alignment \n"
             "and sorting. Since both tasks support multithreading, \n"
             "it is recommended to speicfy a parallel environment (PE) \n"
             "to speed up the process (also see --thread)."
    )
    parser.add_argument(
        'qsub2',
        type=str,
        help="SGE resoruce to request with qsub for the rest of the \n"
             "tasks, which do not support multithreading."
    )
    parser.add_argument(
        'java',
        help='Java resoruce to request for GATK.'
    )
    parser.add_argument(
        'vcf',
        type=str,
        nargs='+',
        help='One or more reference VCF files containing known variant \n'
             'sites (e.g. 1000 Genomes Project).'
    )
    parser.add_argument(
        '--bed',
        metavar='PATH',
        type=str,
        help='BED file.'
    )
    parser.add_argument(
        '--thread',
        metavar='INT',
        type=int,
        default=1,
        help='Number of threads to use (default: 1).'
    )
    parser.add_argument(
        '--platform',
        metavar='TEXT',
        type=str,
        default='Illumina',
        help="Sequencing platform (default: 'Illumina')."
    )
    parser.add_argument(
        '--force',
        action='store_true',
        help='Overwrite the output directory if it already exists.'
    )
    parser.add_argument(
        '--keep',
        action='store_true',
        help='Keep temporary files.'
    )

def main(args):
    if os.path.exists(args.output) and args.force:
        shutil.rmtree(args.output)

    os.mkdir(args.output)
    os.mkdir(f'{args.output}/shell')
    os.mkdir(f'{args.output}/log')
    os.mkdir(f'{args.output}/temp')

    with open(f'{args.output}/command.txt', 'w') as f:
        f.write(' '.join(sys.argv) + '\n')

    df = pd.read_csv(args.manifest)

    if args.keep:
        remove = '# rm'
    else:
        remove = 'rm'

    for i, r in df.iterrows():
        with open(f'{args.output}/shell/S1-{r.Name}.sh', 'w') as f:

            ###########
            # BWA-MEM #
            ###########

            f.write(
f"""#!/bin/bash

# Activate conda environment.
source activate {api.common.conda_env()}

# Get read group information.
first=`zcat {r.Read1} | head -1`
flowcell=`echo "$first" | awk -F " " '{{print $1}}' | awk -F ":" '{{print $3}}'`
barcode=`echo "$first" | awk -F " " '{{print $2}}' | awk -F ":" '{{print $4}}'`
group="@RG\\tID:$flowcell\\tPU:$flowcell.$barcode\\tSM:{r.Name}\\tPL:{args.platform}\\tLB:{r.Name}"

# Align and sort seuqnece reads. Assign read group as well.
bwa mem -M -R $group -t {args.thread} {args.fasta} {r.Read1} {r.Read2} | samtools sort -@ {args.thread} -o {args.output}/temp/{r.Name}.sorted.bam -
""")

        with open(f'{args.output}/shell/S2-{r.Name}.sh', 'w') as f:

            ##################
            # MarkDuplicates #
            ##################

            command1 = 'gatk MarkDuplicates'
            command1 += f' --QUIET'
            command1 += f' --java-options "{args.java}"'
            command1 += f' -I {args.output}/temp/{r.Name}.sorted.bam'
            command1 += f' -O {args.output}/temp/{r.Name}.sorted.markdup.bam'
            command1 += f' -M {args.output}/temp/{r.Name}.metrics'

            ####################
            # BaseRecalibrator #
            ####################

            command2 = 'gatk BaseRecalibrator'
            command2 += f' --QUIET'
            command2 += f' --java-options "{args.java}"'
            command2 += f' -R {args.fasta}'
            command2 += f' -I {args.output}/temp/{r.Name}.sorted.markdup.bam'
            command2 += f' -O {args.output}/temp/{r.Name}.table'
            command2 += ' ' + ' '.join([f'--known-sites {x}' for x in args.vcf])

            if args.bed is not None:
                command2 += f' -L {args.bed}'

            #############
            # ApplyBQSR #
            #############

            command3 = 'gatk ApplyBQSR'
            command3 += f' --QUIET'
            command3 += f' --java-options "{args.java}"'
            command3 += f' -bqsr {args.output}/temp/{r.Name}.table'
            command3 += f' -I {args.output}/temp/{r.Name}.sorted.markdup.bam'
            command3 += f' -O {args.output}/{r.Name}.sorted.markdup.recal.bam'

            if args.bed is not None:
                command3 += f' -L {args.bed}'

            f.write(
f"""#!/bin/bash

# Activate conda environment.
source activate {api.common.conda_env()}

# Mark duplicate reads.
{command1}

# Index BAM file.
samtools index {args.output}/temp/{r.Name}.sorted.markdup.bam

# Build BQSR model.
{command2}

# Apply BQSR model.
{command3}

# Remove temporary files.
{remove} {args.output}/temp/{r.Name}.metrics
{remove} {args.output}/temp/{r.Name}.table
{remove} {args.output}/temp/{r.Name}.sorted.bam
{remove} {args.output}/temp/{r.Name}.sorted.markdup.bam
{remove} {args.output}/temp/{r.Name}.sorted.markdup.bam.bai
""")

    with open(f'{args.output}/shell/qsubme.sh', 'w') as f:
        f.write(
f"""#!/bin/bash

p={args.output}

samples=({" ".join(df.Name)})

for sample in ${{samples[@]}}
do
  qsub {args.qsub1} -S /bin/bash -e $p/log -o $p/log -N S1-$sample $p/shell/S1-$sample.sh
  qsub {args.qsub2} -S /bin/bash -e $p/log -o $p/log -N S2-$sample -hold_jid S1-$sample $p/shell/S2-$sample.sh
done
""")



from .. import api

import matplotlib.pyplot as plt

description = """
Create an oncoplot with a MAF file.

The format of output image (PDF/PNG/JPEG/SVG) will be automatically
determined by the output file's extension.
"""

epilog = f"""
[Example] Output a PNG file:
  $ fuc {api.common._script_name()} in.maf out.png

[Example] Output a PDF file:
  $ fuc {api.common._script_name()} in.maf out.pdf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Create an oncoplot with a MAF file.',
    )
    parser.add_argument(
        'maf',
        help='MAF file.'
    )
    parser.add_argument(
        'out',
        help='Output image file.'
    )
    parser.add_argument(
        '--count',
        metavar='INT',
        type=int,
        default=10,
        help='Number of top mutated genes to display (default: 10).'
    )
    parser.add_argument(
        '--figsize',
        metavar='FLOAT',
        type=float,
        default=[15, 10],
        nargs=2,
        help='Width, height in inches (default: [15, 10]).'
    )
    parser.add_argument(
        '--label_fontsize',
        metavar='FLOAT',
        type=float,
        default=15,
        help='Font size of labels (default: 15).'
    )
    parser.add_argument(
        '--ticklabels_fontsize',
        metavar='FLOAT',
        type=float,
        default=15,
        help='Font size of tick labels (default: 15).'
    )
    parser.add_argument(
        '--legend_fontsize',
        metavar='FLOAT',
        type=float,
        default=15,
        help='Font size of legend texts (default: 15).'
    )

def main(args):
    mf = api.pymaf.MafFrame.from_file(args.maf)
    mf.plot_oncoplot(
        count=args.count,
        figsize=args.figsize,
        label_fontsize=args.label_fontsize,
        ticklabels_fontsize=args.ticklabels_fontsize,
        legend_fontsize=args.legend_fontsize
    )
    plt.savefig(args.out)



import sys
import subprocess

from .. import api

description = """
Count sequence reads in FASTQ files.
"""

epilog = f"""
[Example] When the input is a FASTQ file:
  $ fuc {api.common._script_name()} in1.fastq in2.fastq

[Example] When the input is stdin:
  $ cat fastq.list | fuc {api.common._script_name()}
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Count sequence reads in FASTQ files.',
    )
    parser.add_argument(
        'fastq',
        nargs='*',
        help='FASTQ files (zipped or unzipped) (default: stdin).'
    )

def main(args):
    if args.fastq:
        fastqs = args.fastq
    elif not sys.stdin.isatty():
        fastqs = sys.stdin.read().rstrip('\n').split('\n')
    else:
        raise ValueError('No input files detected.')

    for fastq in fastqs:
        if fastq.endswith('.gz'):
            cat = 'zcat'
        else:
            cat = 'cat'
        command = f'echo $({cat} < {fastq} | wc -l) / 4 | bc'
        subprocess.run(command, shell=True)



"""
The pybam submodule is designed for working with sequence alignment files
(SAM/BAM/CRAM). It essentially wraps the `pysam
<https://pysam.readthedocs.io/en/latest/api.html>`_ package to allow fast
computation and easy manipulation. If you are mainly interested in working
with depth of coverage data, please check out the pycov submodule which is
specifically designed for the task.
"""

from . import common

import pysam
import pandas as pd

def tag_sm(fn):
    """
    Extract SM tags (sample names) from a BAM file.

    Parameters
    ----------
    fn : str
        BAM file.

    Returns
    -------
    list
        List of SM tags.

    Examples
    --------

    >>> from fuc import pybam
    >>> pybam.tag_sm('NA19920.bam')
    ['NA19920']
    """
    lines = pysam.view('-H', fn, '--no-PG').strip().split('\n')
    tags = []
    for line in lines:
        fields = line.split('\t')
        if fields[0] == '@RG':
            for field in fields:
                if 'SM:' in field:
                    tags.append(field.replace('SM:', ''))
    return list(set(tags))

def tag_sn(fn):
    """
    Extract SN tags (contig names) from a BAM file.

    Parameters
    ----------
    fn : str
        BAM file.

    Returns
    -------
    list
        List of SN tags.

    Examples
    --------

    >>> from fuc import pybam
    >>> pybam.tag_sn('NA19920.bam')
    ['chr3', 'chr15', 'chrY', 'chr19', 'chr22', 'chr5', 'chr18', 'chr14', 'chr11', 'chr20', 'chr21', 'chr16', 'chr10', 'chr13', 'chr9', 'chr2', 'chr17', 'chr12', 'chr6', 'chrM', 'chrX', 'chr4', 'chr8', 'chr1', 'chr7']
    """
    lines = pysam.view('-H', fn, '--no-PG').strip().split('\n')
    tags = []
    for line in lines:
        fields = line.split('\t')
        if fields[0] == '@SQ':
            for field in fields:
                if 'SN:' in field:
                    tags.append(field.replace('SN:', ''))
    return list(set(tags))

def has_chr_prefix(fn):
    """
    Return True if contigs have the (annoying) 'chr' string.

    Parameters
    ----------
    fn : str
        BAM file.

    Returns
    -------
    bool
        Whether the 'chr' string is found.
    """
    contigs = tag_sn(fn)
    for contig in contigs:
        if 'chr' in contig:
            return True
    return False

def count_allelic_depth(bam, sites):
    """
    Count allelic depth for specified sites.

    Parameters
    ----------
    bam : str
        BAM file.
    sites : str or list
        Genomic site or list of sites. Each site should consist of chromosome
        and 1-based position in the format that can be recognized by
        :meth:`common.parse_variant` (e.g. '22-42127941').

    Returns
    -------
    pandas.DataFrame
        DataFrame containing allelic depth.

    Examples
    --------

    >>> from fuc import pybam
    >>> pybam.count_allelic_depth('in.bam', ['19-41510048', '19-41510053', '19-41510062'])
      Chromosome  Position  Total    A  C    G    T  N  DEL  INS
    0         19  41510048    119  106  7    4    0  0    2    0
    1         19  41510053    120    1  2    0  116  0    0    1
    2         19  41510062    115    0  0  115    0  0    0    0
    """
    if isinstance(sites, str):
        sites = [sites]

    has_prefix = has_chr_prefix(bam)

    f = pysam.AlignmentFile(bam, 'rb')

    rows = []

    for site in sites:
        chrom, pos, _, _ = common.parse_variant(site)

        if has_prefix:
            if 'chr' not in chrom:
                formatted_chrom = 'chr' + chrom
            else:
                formatted_chrom = chrom
        else:
            if 'chr' in chrom:
                formatted_chrom = chrom.replace('chr', '')
            else:
                formatted_chrom = chrom
        row = {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'N': 0 , 'DEL': 0, 'INS': 0}
        kwargs = dict(
            min_base_quality=0,
            ignore_overlaps=False,
            ignore_orphans=False,
            truncate=True
        )
        for i, col in enumerate(f.pileup(formatted_chrom, pos-2, pos, **kwargs)):
            for read in col.pileups:
                if i == 0:
                    if read.indel < 0:
                        row['DEL'] += 1
                    elif read.indel > 0:
                        row['INS'] += 1
                    else:
                        continue
                else:
                    if read.is_del:
                        continue
                    allele = read.alignment.query_sequence[read.query_position]
                    row[allele] += 1

        data = list(row.values())
        rows.append([chrom, pos, sum(data)] + data)

    f.close()

    return pd.DataFrame(rows, columns=['Chromosome', 'Position', 'Total']+list(row))



import os
import sys
import shutil

from .. import api

import pandas as pd

description = f"""
Pipeline for constructing a panel of normals (PoN).

Dependencies:
  - GATK: Required for constructing PoN.

Manifest columns:
  - BAM: Path to recalibrated BAM file.
"""

epilog = f"""
[Example] Specify queue:
  $ fuc {api.common._script_name()} \\
  manifest.csv \\
  ref.fa \\
  output_dir \\
  "-q queue_name" \\
  "-Xmx15g -Xms15g"

[Example] Specify nodes:
  $ fuc {api.common._script_name()} \\
  manifest.csv \\
  ref.fa \\
  output_dir \\
  "-l h='node_A|node_B'" \\
  "-Xmx15g -Xms15g"
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Pipeline for constructing a panel of normals (PoN).',
    )
    parser.add_argument(
        'manifest',
        help='Sample manifest CSV file.'
    )
    parser.add_argument(
        'fasta',
        help='Reference FASTA file.'
    )
    parser.add_argument(
        'output',
        type=os.path.abspath,
        help='Output directory.'
    )
    parser.add_argument(
        'qsub',
        type=str,
        help='SGE resoruce to request for qsub.'
    )
    parser.add_argument(
        'java',
        help='Java resoruce to request for GATK.'
    )
    parser.add_argument(
        '--bed',
        metavar='PATH',
        type=str,
        help='BED file.'
    )
    parser.add_argument(
        '--force',
        action='store_true',
        help='Overwrite the output directory if it already exists.'
    )
    parser.add_argument(
        '--keep',
        action='store_true',
        help='Keep temporary files.'
    )

def main(args):
    if os.path.exists(args.output) and args.force:
        shutil.rmtree(args.output)

    os.mkdir(args.output)
    os.mkdir(f'{args.output}/shell')
    os.mkdir(f'{args.output}/log')
    os.mkdir(f'{args.output}/temp')

    with open(f'{args.output}/command.txt', 'w') as f:
        f.write(' '.join(sys.argv) + '\n')

    df = pd.read_csv(args.manifest)

    if args.keep:
        remove = '# rm'
    else:
        remove = 'rm'

    basenames = []

    for i, r in df.iterrows():
        basename = os.path.basename(r.BAM)
        basenames.append(basename)

        with open(f'{args.output}/shell/S1-{basename}.sh', 'w') as f:

            ###########
            # Mutect2 #
            ###########

            command = 'gatk Mutect2'
            command += f' --QUIET'
            command += f' --java-options "{args.java}"'
            command += f' -R {args.fasta}'
            command += f' -I {r.BAM}'
            command += f' -max-mnp-distance 0'
            command += f' -O {args.output}/temp/{basename}.vcf'

            if args.bed is not None:
                command += f' -L {args.bed}'

            f.write(
f"""#!/bin/bash

# Activate conda environment.
source activate {api.common.conda_env()}

# Run Mutect2 in tumor-only mode for each normal sample.
{command}
""")

    chroms = [str(i) for i in range(1, 23)] + ['X', 'Y']

    if api.pybam.has_chr_prefix(df.BAM[0]):
        chroms = ['chr' + x for x in chroms]

    for chrom in chroms:
        with open(f'{args.output}/shell/S2-{chrom}.sh', 'w') as f:

            ####################
            # GenomicsDBImport #
            ####################

            command1 = 'gatk GenomicsDBImport'
            command1 += f' --QUIET'
            command1 += f' --java-options "{args.java}"'
            command1 += f' -R {args.fasta}'
            command1 += f' -L {chrom}'
            command1 += f' --genomicsdb-workspace-path {args.output}/temp/db-{chrom}'
            command1 += ' ' + ' '.join([f'-V {args.output}/temp/{x}.vcf' for x in basenames])

            ###############################
            # CreateSomaticPanelOfNormals #
            ###############################

            command2 = 'gatk CreateSomaticPanelOfNormals'
            command2 += f' --QUIET'
            command2 += f' --java-options "{args.java}"'
            command2 += f' -R {args.fasta}'
            command2 += f' -V gendb://{args.output}/temp/db-{chrom}'
            command2 += f' -O {args.output}/temp/{chrom}.pon.vcf'

            f.write(
f"""#!/bin/bash

# Activate conda environment.
source activate {api.common.conda_env()}

# Create a GenomicsDB from the normal Mutect2 calls.
{command1}

# Combine the normal calls.
{command2}
""")

    with open(f'{args.output}/shell/S3.sh', 'w') as f:

        ##############
        # GatherVcfs #
        ##############

        command = 'gatk GatherVcfs'
        command += f' --QUIET'
        command += f' --java-options "{args.java}"'
        command += f' -O {args.output}/merged.pon.vcf'
        command += ' ' + ' '.join([f'-I {args.output}/temp/{x}.pon.vcf' for x in chroms])

        f.write(
f"""#!/bin/bash

# Activate conda environment.
source activate {api.common.conda_env()}

# Merge VCF files.
{command}

# Remove temporary files.
{remove} -r {args.output}/temp/*
""")

    with open(f'{args.output}/shell/qsubme.sh', 'w') as f:
        f.write(
f"""#!/bin/bash

p={args.output}

samples=({" ".join(basenames)})

chroms=({" ".join(chroms)})

for sample in ${{samples[@]}}
do
  qsub {args.qsub} -S /bin/bash -e $p/log -o $p/log -N S1 $p/shell/S1-$sample.sh
done

for chrom in ${{chroms[@]}}
do
  qsub {args.qsub} -S /bin/bash -e $p/log -o $p/log -N S2 -hold_jid S1 $p/shell/S2-$chrom.sh
done

qsub {args.qsub} -S /bin/bash -e $p/log -o $p/log -N S3 -hold_jid S2 $p/shell/S3.sh
""")



import gzip

from .. import api

description = f"""
Compute top unknown barcodes using undertermined FASTQ from bcl2fastq.

This command will compute top unknown barcodes using undertermined FASTQ from
the bcl2fastq or bcl2fastq2 prograrm.
"""

epilog = f"""
[Example] Compute top unknown barcodes:
  $ fuc {api.common._script_name()} Undetermined_S0_R1_001.fastq.gz
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Compute top unknown barcodes using undertermined FASTQ '
             'from bcl2fastq.',
    )
    parser.add_argument(
        'fastq',
        help='Undertermined FASTQ (zipped or unzipped).'
    )
    parser.add_argument(
        '--count',
        metavar='INT',
        type=int,
        default=30,
        help='Number of top unknown barcodes to return (default: 30).'
    )

def main(args):
    if args.fastq.endswith('.gz'):
        f = gzip.open(args.fastq, 'rt')
    else:
        f = open(args.fastq)

    data = {}

    for line in f:
        if not line.startswith('@'):
            continue
        fields = line.strip().split(':')
        name = fields[-1]
        if name not in data:
            data[name] = 0
        data[name] += 1

    f.close()

    data = dict(sorted(data.items(), key=lambda x: x[1], reverse=True))

    for i, name in enumerate(data):
        if i < 30:
            print(name, data[name])



import sys

from .. import api

import pysam

description = """
Print the header of a SAM/BAM/CRAM file.
"""

epilog = f"""
[Example] Print the header of a BAM file:
  $ fuc {api.common._script_name()} in.bam
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Print the header of a SAM/BAM/CRAM file.',
    )
    parser.add_argument(
        'bam',
        help='Alignment file.'
    )

def main(args):
    header = pysam.view('-H', args.bam, '--no-PG')
    sys.stdout.write(header)



import sys

from .. import api

description = f"""
Filter a VCF file by annotations from Ensembl VEP.
"""

epilog = f"""
[Example] Select variants in the TP53 gene:
  $ fuc {api.common._script_name()} in.vcf "SYMBOL == 'TP53'" > out.vcf

[Example] Exclude variants from the TP53 gene:
  $ fuc {api.common._script_name()} in.vcf "SYMBOL != 'TP53'" > out.vcf

[Example] Same as above:
  $ fuc {api.common._script_name()} in.vcf "SYMBOL == 'TP53'" --opposite > out.vcf

[Example] Select splice donor or stop-gain variants:
  $ fuc {api.common._script_name()} in.vcf \\
  "Consequence in ['splice_donor_variant', 'stop_gained']" > out.vcf

[Example] Build a complex query to select specific variants:
  $ fuc {api.common._script_name()} in.vcf \\
  "(SYMBOL == 'TP53') and (Consequence.str.contains('stop_gained'))" > out.vcf

[Example] Select variants whose gnomAD AF is less than 0.001:
  $ fuc {api.common._script_name()} in.vcf "gnomAD_AF < 0.001" > out.vcf

[Example] Variants without AF data will be treated as having AF of 0:
  $ fuc {api.common._script_name()} in.vcf "gnomAD_AF < 0.001" --as_zero > out.vcf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Filter a VCF file by annotations from Ensembl VEP.',
    )
    parser.add_argument(
        'vcf',
        help='VCF file annotated by Ensembl VEP (zipped or unzipped).'
    )
    parser.add_argument(
        'expr',
        help='Query expression to evaluate.'
    )
    parser.add_argument(
        '--opposite',
        action='store_true',
        help="Use this flag to return only records that don't \n"
             "meet the said criteria."
    )
    parser.add_argument(
        '--as_zero',
        action='store_true',
        help='Use this flag to treat missing values as zero instead of NaN.'
    )

def main(args):
    vf = api.pyvcf.VcfFrame.from_file(args.vcf)
    filtered_vf = api.pyvep.filter_query(vf, args.expr,
        opposite=args.opposite, as_zero=args.as_zero)
    sys.stdout.write(filtered_vf.to_string())



import sys

from .. import api

description = """
Convert a VCF file to a BED file.
"""

epilog = f"""
[Example] Convert VCF to BED:
  $ fuc {api.common._script_name()} in.vcf > out.bed
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Convert a VCF file to a BED file.',
    )
    parser.add_argument(
        'vcf',
        help='VCF file (zipped or unzipped).'
    )

def main(args):
    vf = api.pyvcf.VcfFrame.from_file(args.vcf)
    bf = vf.to_bed()
    sys.stdout.write(bf.to_string())



from .. import api

import matplotlib.pyplot as plt

description = """
Create a summary plot with a MAF file.

The format of output image (PDF/PNG/JPEG/SVG) will be automatically
determined by the output file's extension.
"""

epilog = f"""
[Example] Output a PNG file:
  $ fuc {api.common._script_name()} in.maf out.png

[Example] Output a PNG file:
  $ fuc {api.common._script_name()} in.maf out.pdf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Create a summary plot with a MAF file.',
    )
    parser.add_argument(
        'maf',
        help='MAF file.'
    )
    parser.add_argument(
        'out',
        help='Output image file.'
    )
    parser.add_argument(
        '--figsize',
        metavar='FLOAT',
        type=float,
        default=[15, 10],
        nargs=2,
        help='width, height in inches (default: [15, 10])'
    )
    parser.add_argument(
        '--title_fontsize',
        metavar='FLOAT',
        type=float,
        default=16,
        help='font size of subplot titles (default: 16)'
    )
    parser.add_argument(
        '--ticklabels_fontsize',
        metavar='FLOAT',
        type=float,
        default=12,
        help='font size of tick labels (default: 12)'
    )
    parser.add_argument(
        '--legend_fontsize',
        metavar='FLOAT',
        type=float,
        default=12,
        help='font size of legend texts (default: 12)'
    )

def main(args):
    mf = api.pymaf.MafFrame.from_file(args.maf)
    mf.plot_summary(
        figsize=args.figsize,
        title_fontsize=args.title_fontsize,
        ticklabels_fontsize=args.ticklabels_fontsize,
        legend_fontsize=args.legend_fontsize
    )
    plt.savefig(args.out)



import sys

from .. import api

description = """
Find the intersection of BED files.
"""

epilog = f"""
[Example] Find the intersection of three BED files:
  $ fuc {api.common._script_name()} in1.bed in2.bed in3.bed > out.bed
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Find the intersection of BED files.',
    )
    parser.add_argument('bed', help='BED files.', nargs='+')

def main(args):
    bfs = [api.pybed.BedFrame.from_file(x) for x in args.bed]
    final_bf = bfs[0]
    for bf in bfs[1:]:
        final_bf = final_bf.intersect(bf)
    sys.stdout.write(final_bf.to_string())



import sys

from .. import api

import pysam

description = """
Index a GFF/BED/SAM/VCF file with Tabix.

The Tabix program is used to index a TAB-delimited genome position file
(GFF/BED/SAM/VCF) and create an index file (.tbi). The input data file must
be position sorted and compressed by bgzip.
"""

epilog = f"""
[Example] Index a GFF file:
  $ fuc {api.common._script_name()} in.gff.gz

[Example] Index a BED file:
  $ fuc {api.common._script_name()} in.bed.gz

[Example] Index a SAM file:
  $ fuc {api.common._script_name()} in.sam.gz

[Example] Index a VCF file.
  $ fuc {api.common._script_name()} in.vcf.gz
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Index a GFF/BED/SAM/VCF file with Tabix.',
    )
    parser.add_argument(
        'file',
        help='File to be indexed.'
    )
    parser.add_argument(
        '--force',
        action='store_true',
        help='Force to overwrite the index file if it is present.'
    )

def main(args):
    if '.vcf' in args.file:
        preset = 'vcf'
    elif '.sam' in args.file:
        preset = 'sam'
    elif '.gff' in args.file:
        preset = 'gff'
    elif '.bed' in args.file:
        preset = 'bed'
    else:
        raise ValueError('Unsupported file format')

    pysam.tabix_index(args.file, preset=preset, force=args.force)



import sys
import tempfile

from .. import api

import pysam

description = f"""
Rename the sample in a SAM/BAM/CRAM file.
"""

epilog = f"""
[Example] Write a new BAM file after renaming:
  $ fuc {api.common._script_name()} in.bam NA12878 > out.bam
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Rename the sample in a SAM/BAM/CRAM file.',
    )
    parser.add_argument('bam', help='Alignment file.')
    parser.add_argument('name', help='New sample name.')

def main(args):
    # Detect the input format.
    try:
        with open(args.bam, 'r') as f:
            for line in f:
                is_sam = True
                break
    except UnicodeDecodeError:
        is_sam = False

    # Rename the sample(s).
    old_lines = pysam.view('-H', args.bam, '--no-PG').strip().split('\n')
    new_lines = []
    for old_line in old_lines:
        old_fields = old_line.split('\t')
        if old_fields[0] != '@RG':
            new_lines.append(old_line)
            continue
        new_fields = []
        for old_field in old_fields:
            if 'SM:' not in old_field:
                new_fields.append(old_field)
                continue
            new_fields.append(f'SM:{args.name}')
        new_lines.append('\t'.join(new_fields))

    # Write the output file.
    if is_sam:
        for new_line in new_lines:
            sys.stdout.write(new_line + '\n')
        alignments = pysam.view(args.bam, '--no-PG')
        sys.stdout.write(alignments)
    else:
        with tempfile.TemporaryDirectory() as t:
            with open(f'{t}/header.sam', 'w') as f:
                f.write('\n'.join(new_lines))
            alignments = pysam.reheader(f'{t}/header.sam',
                                        args.bam,
                                        '--no-PG')
            sys.stdout.buffer.write(alignments)



import sys

from .. import api

from Bio import bgzf

description = """
Write a BGZF compressed file.

BGZF (Blocked GNU Zip Format) is a modified form of gzip compression which
can be applied to any file format to provide compression with efficient
random access. In addition to being required for random access to and writing
of BAM files, the BGZF format can also be used for most of the sequence data
formats (e.g. FASTA, FASTQ, GenBank, VCF, MAF).
"""

epilog = f"""
[Example] When the input is a VCF file:
  $ fuc {api.common._script_name()} in.vcf > out.vcf.gz

[Example] When the input is stdin:
  $ cat in.vcf | fuc {api.common._script_name()} > out.vcf.gz
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Write a BGZF compressed file.',
    )
    parser.add_argument(
        'file',
        nargs='*',
        help='File to be compressed (default: stdin).'
    )

def main(args):
    if args.file:
        with open(args.file[0]) as f:
            data = f.read()
    elif not sys.stdin.isatty():
        data = sys.stdin.read()
    else:
        raise ValueError('No input data detected')

    w = bgzf.BgzfWriter(fileobj=sys.stdout.buffer)
    w.write(data)
    w.close()



import sys

from .. import api

description = f"""
Merge two or more VCF files.
"""

epilog = f"""
[Example] Merge multiple VCF files:
  $ fuc {api.common._script_name()} 1.vcf 2.vcf 3.vcf > merged.vcf

[Example] Keep the GT, AD, DP fields:
  $ fuc {api.common._script_name()} 1.vcf 2.vcf --format GT:AD:DP > merged.vcf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Merge two or more VCF files.',
    )
    parser.add_argument(
        'vcf_files',
        help='VCF files (zipped or unzipped).',
        nargs='+'
    )
    parser.add_argument(
        '--how',
        metavar='TEXT',
        default='inner',
        help="Type of merge as defined in pandas.DataFrame.merge \n"
             "(default: 'inner')."
    )
    parser.add_argument(
        '--format',
        metavar='TEXT',
        default='GT',
        help="FORMAT subfields to be retained (e.g. 'GT:AD:DP') \n"
             "(default: 'GT')."
    )
    parser.add_argument(
        '--sort',
        action='store_false',
        help='Use this flag to turn off sorting of records \n'
             '(default: True).'
    )
    parser.add_argument(
        '--collapse',
        action='store_true',
        help='Use this flag to collapse duplicate records \n'
             '(default: False).'
    )

def main(args):
    vfs = [api.pyvcf.VcfFrame.from_file(x) for x in args.vcf_files]
    merged_vf = api.pyvcf.merge(vfs, format=args.format, how=args.how,
        sort=args.sort, collapse=args.collapse)
    sys.stdout.write(merged_vf.to_string())



import sys

from .. import api

import pysam

description = """
Compute read depth from SAM/BAM/CRAM files.

By default, the command will count all reads within the alignment files. You
can specify regions of interest with --bed or --region. When you do this, pay
close attention to the 'chr' string in contig names (e.g. 'chr1' vs. '1').
Note also that --region requires the input files be indexed.
"""

epilog = f"""
[Example] To specify regions with a BED file:
  $ fuc {api.common._script_name()} \\
  --bam 1.bam 2.bam \\
  --bed in.bed > out.tsv

[Example] To specify regions manually:
  $ fuc {api.common._script_name()} \\
  --fn bam.list \\
  --region chr1:100-200 > out.tsv
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Compute read depth from SAM/BAM/CRAM files.',
    )
    parser.add_argument(
        '--bam',
        metavar='PATH',
        nargs='+',
        help='One or more alignment files. Cannot be used with --fn.'
    )
    parser.add_argument(
        '--fn',
        metavar='PATH',
        help='File containing one alignment file per line. Cannot \n'
             'be used with --bam.'
    )
    parser.add_argument(
        '--bed',
        metavar='PATH',
        help='BED file. Cannot be used with --region.'
    )
    parser.add_argument(
        '--region',
        metavar='TEXT',
        help="Target region ('chrom:start-end'). Cannot be used \n"
             "with --bed."
    )
    parser.add_argument(
        '--zero',
        action='store_true',
        help='Output all positions including those with zero depth.'
    )

def main(args):
    cf = api.pycov.CovFrame.from_bam(
        bam=args.bam, fn=args.fn, bed=args.bed, region=args.region,
        zero=args.zero
    )
    sys.stdout.write(cf.to_string())



import os
import sys
import shutil

from .. import api

import pandas as pd

description = """
Pipeline for somatic short variant discovery.

External dependencies:
  - SGE: Required for job submission (i.e. qsub).
  - GATK: Required for variant calling (i.e. Mutect2) and filtration.

Manifest columns:
  - Tumor: Recalibrated BAM file for tumor.
  - Normal: Recalibrated BAM file for matched normal.
"""

epilog = f"""
[Example] Specify queue:
  $ fuc {api.common._script_name()} \\
  manifest.csv \\
  ref.fa \\
  output_dir \\
  pon.vcf \\
  germline.vcf \\
  "-q queue_name" \\
  "-Xmx15g -Xms15g"

[Example] Specify nodes:
  $ fuc {api.common._script_name()} \\
  manifest.csv \\
  ref.fa \\
  output_dir \\
  pon.vcf \\
  germline.vcf \\
  "-l h='node_A|node_B'" \\
  "-Xmx15g -Xms15g" \\
  --bed in.bed
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        help='Pipeline for somatic short variant discovery.',
        description=description,
    )
    parser.add_argument(
        'manifest',
        help='Sample manifest CSV file.'
    )
    parser.add_argument(
        'fasta',
        help='Reference FASTA file.'
    )
    parser.add_argument(
        'output',
        type=os.path.abspath,
        help='Output directory.'
    )
    parser.add_argument(
        'pon',
        help='PoN VCF file.'
    )
    parser.add_argument(
        'germline',
        help='Germline VCF file.'
    )
    parser.add_argument(
        'qsub',
        type=str,
        help='SGE resoruce to request for qsub.'
    )
    parser.add_argument(
        'java',
        type=str,
        help='Java resoruce to request for GATK.'
    )
    parser.add_argument(
        '--bed',
        metavar='PATH',
        type=str,
        help='BED file.'
    )
    parser.add_argument(
        '--force',
        action='store_true',
        help='Overwrite the output directory if it already exists.'
    )
    parser.add_argument(
        '--keep',
        action='store_true',
        help='Keep temporary files.'
    )

def main(args):
    if os.path.exists(args.output) and args.force:
        shutil.rmtree(args.output)

    os.mkdir(args.output)
    os.mkdir(f'{args.output}/shell')
    os.mkdir(f'{args.output}/log')
    os.mkdir(f'{args.output}/temp')

    with open(f'{args.output}/command.txt', 'w') as f:
        f.write(' '.join(sys.argv) + '\n')

    df = pd.read_csv(args.manifest)

    if args.keep:
        remove = '# rm'
    else:
        remove = 'rm'

    basenames = []

    for i, r in df.iterrows():
        basename = api.pybam.tag_sm(r.Tumor)[0]
        basenames.append(basename)

        with open(f'{args.output}/shell/S1-{basename}.sh', 'w') as f:

            ###########
            # Mutect2 #
            ###########

            command1 = 'gatk Mutect2'
            command1 += f' --QUIET'
            command1 += f' --java-options "{args.java}"'
            command1 += f' -R {args.fasta}'
            command1 += f' -pon {args.pon}'
            command1 += f' --germline-resource {args.germline}'
            command1 += f' --f1r2-tar-gz {args.output}/temp/{basename}.f1r2.tar.gz'
            command1 += f' -I {r.Tumor}'
            command1 += f' -I {r.Normal}'
            command1 += f' -normal {api.pybam.tag_sm(r.Normal)[0]}'
            command1 += f' -O {args.output}/temp/{basename}.raw.vcf'

            if args.bed is not None:
                command1 += f' -L {args.bed}'

            ######################
            # GetPileupSummaries #
            ######################

            command2 = 'gatk GetPileupSummaries'
            command2 += f' --QUIET'
            command2 += f' --java-options "{args.java}"'
            command2 += f' -I {r.Tumor}'
            command2 += f' -V {args.germline}'
            command2 += f' -L {args.output}/temp/{basename}.raw.vcf'
            command2 += f' -O {args.output}/temp/{basename}.tumor-pileups.table'

            ######################
            # GetPileupSummaries #
            ######################

            command3 = 'gatk GetPileupSummaries'
            command3 += f' --QUIET'
            command3 += f' --java-options "{args.java}"'
            command3 += f' -I {r.Normal}'
            command3 += f' -V {args.germline}'
            command3 += f' -L {args.output}/temp/{basename}.raw.vcf'
            command3 += f' -O {args.output}/temp/{basename}.normal-pileups.table'

            ##########################
            # CalculateContamination #
            ##########################

            command4 = 'gatk CalculateContamination'
            command4 += f' --QUIET'
            command4 += f' --java-options "{args.java}"'
            command4 += f' -I {args.output}/temp/{basename}.tumor-pileups.table'
            command4 += f' -matched {args.output}/temp/{basename}.normal-pileups.table'
            command4 += f' -O {args.output}/temp/{basename}.contamination.table'
            command4 += f' -segments {args.output}/temp/{basename}.segments.tsv'

            #############################
            # LearnReadOrientationModel #
            #############################

            command5 = 'gatk LearnReadOrientationModel'
            command5 += f' --QUIET'
            command5 += f' --java-options "{args.java}"'
            command5 += f' -I {args.output}/temp/{basename}.f1r2.tar.gz'
            command5 += f' -O {args.output}/temp/{basename}.artifact-prior.tar.gz'

            #####################
            # FilterMutectCalls #
            #####################

            command6 = 'gatk FilterMutectCalls'
            command6 += f' --QUIET'
            command6 += f' --java-options "{args.java}"'
            command6 += f' -R {args.fasta}'
            command6 += f' -V {args.output}/temp/{basename}.raw.vcf'
            command6 += f' --contamination-table {args.output}/temp/{basename}.contamination.table'
            command6 += f' --tumor-segmentation {args.output}/temp/{basename}.segments.tsv'
            command6 += f' -ob-priors {args.output}/temp/{basename}.artifact-prior.tar.gz'
            command6 += f' -O {args.output}/{basename}.filtered.vcf'
            command6 += f' --filtering-stats {args.output}/temp/{basename}.filtered.vcf.filteringStats.tsv'

            f.write(
f"""#!/bin/bash

## Activate conda environment.
source activate {api.common.conda_env()}

## Call candidate variants.
{command1}

## Get pileup summary for tumor.
{command2}

## Get pileup summary for normal.
{command3}

## Calculate contamination.
{command4}

## Construct orientation bias model.
{command5}

## Filter variants.
{command6}

# Remove temporary files.
{remove} {args.output}/temp/{basename}.f1r2.tar.gz
{remove} {args.output}/temp/{basename}.normal-pileups.table
{remove} {args.output}/temp/{basename}.tumor-pileups.table
{remove} {args.output}/temp/{basename}.raw.vcf
{remove} {args.output}/temp/{basename}.contamination.table
{remove} {args.output}/temp/{basename}.segments.tsv
{remove} {args.output}/temp/{basename}.artifact-prior.tar.gz
{remove} {args.output}/temp/{basename}.filtered.vcf.filteringStats.tsv
""")

    with open(f'{args.output}/shell/qsubme.sh', 'w') as f:
        f.write(
f"""#!/bin/bash

p={args.output}

samples=({" ".join(basenames)})

for sample in ${{samples[@]}}
do
  qsub {args.qsub} -S /bin/bash -e $p/log -o $p/log -N S1 $p/shell/S1-$sample.sh
done
""")



"""
The pyvep submodule is designed for parsing VCF annotation data from the
`Ensembl VEP <https://asia.ensembl.org/info/docs/tools/vep/index.html>`_
program. It should be used with ``pyvcf.VcfFrame``.

The input VCF should already contain functional annotation data from
Ensembl VEP. The recommended method is Ensembl VEP's
`web interface <https://asia.ensembl.org/Tools/VEP>`_ with
“RefSeq transcripts” as the transcript database and the filtering option
“Show one selected consequence per variant”.

A typical VEP-annotated VCF file contains many fields ranging from gene
symbol to protein change. However, most of the analysis in pyvep uses the
following fields:

+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| No. | Name               | Description         | Examples                                                                                                 |
+=====+====================+=====================+==========================================================================================================+
| 1   | Allele             | Variant allele      | 'C', 'T', 'CG'                                                                                           |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 2   | Consequence        | Consequence type    | 'missense_variant', 'stop_gained'                                                                        |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 3   | IMPACT             | Consequence impact  | 'MODERATE', 'HIGH'                                                                                       |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 4   | SYMBOL             | Gene symbol         | 'TP53', 'APC'                                                                                            |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 5   | Gene               | Ensembl ID          | 7157, 55294                                                                                              |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 6   | BIOTYPE            | Transcript biotype  | 'protein_coding', 'pseudogene', 'miRNA'                                                                  |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 7   | EXON               | Exon number         | '8/17', '17/17'                                                                                          |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 8   | Protein_position   | Amino acid position | 234, 1510                                                                                                |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 9   | Amino_acids        | Amino acid changes  | 'R/\*', 'A/X', 'V/A'                                                                                     |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 10  | Existing_variation | Variant identifier  | 'rs17851045', 'rs786201856&COSV57325157'                                                                 |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 11  | STRAND             | Genomic strand      | 1, -1                                                                                                    |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 12  | SIFT               | SIFT prediction     | 'tolerated(0.6)', 'tolerated_low_confidence(0.37)', 'deleterious_low_confidence(0.03)', 'deleterious(0)' |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 13  | PolyPhen           | PolyPhen prediction | 'benign(0.121)', 'possibly_damaging(0.459)', 'probably_damaging(0.999)'                                  |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
| 14  | CLIN_SIG           | ClinVar prediction  | 'pathogenic', 'likely_pathogenic&pathogenic'                                                             |
+-----+--------------------+---------------------+----------------------------------------------------------------------------------------------------------+
"""

import re
import pandas as pd
import numpy as np
from . import pyvcf

# https://m.ensembl.org/info/docs/tools/vep/vep_formats.html

DATA_TYPES = {
    'Allele': 'str',
    'Consequence': 'str',
    'IMPACT': 'str',
    'SYMBOL': 'str',
    'Gene': 'str',
    'Feature_type': 'str',
    'Feature': 'str',
    'BIOTYPE': 'str',
    'EXON': 'str',
    'INTRON': 'str',
    'HGVSc': 'str',
    'HGVSp': 'str',
    'cDNA_position': 'str',
    'CDS_position': 'str',
    'Protein_position': 'str',
    'Amino_acids': 'str',
    'Codons': 'str',
    'Existing_variation': 'str',
    'DISTANCE': 'str',
    'STRAND': 'str',
    'FLAGS': 'str',
    'SYMBOL_SOURCE': 'str',
    'HGNC_ID': 'str',
    'MANE_SELECT': 'str',
    'MANE_PLUS_CLINICAL': 'str',
    'TSL': 'str',
    'APPRIS': 'str',
    'REFSEQ_MATCH': 'str',
    'REFSEQ_OFFSET': 'str',
    'GIVEN_REF': 'str',
    'USED_REF': 'str',
    'BAM_EDIT': 'str',
    'SIFT': 'str',
    'PolyPhen': 'str',
    'AF': 'float64',
    'AFR_AF': 'float64',
    'AMR_AF': 'float64',
    'EAS_AF': 'float64',
    'EUR_AF': 'float64',
    'SAS_AF': 'float64',
    'AA_AF': 'float64',
    'EA_AF': 'float64',
    'gnomAD_AF': 'float64',
    'gnomAD_AFR_AF': 'float64',
    'gnomAD_AMR_AF': 'float64',
    'gnomAD_ASJ_AF': 'float64',
    'gnomAD_EAS_AF': 'float64',
    'gnomAD_FIN_AF': 'float64',
    'gnomAD_NFE_AF': 'float64',
    'gnomAD_OTH_AF': 'float64',
    'gnomAD_SAS_AF': 'float64',
    'CLIN_SIG': 'str',
    'SOMATIC': 'str',
    'PHENO': 'str',
    'PUBMED': 'str',
    'MOTIF_NAME': 'str',
    'MOTIF_POS': 'str',
    'HIGH_INF_POS': 'str',
    'MOTIF_SCORE_CHANGE': 'str',
    'TRANSCRIPTION_FACTORS': 'str',
}

SEVERITIY = [
    'transcript_ablation',
    'splice_acceptor_variant',
    'splice_donor_variant',
    'stop_gained',
    'frameshift_variant',
    'stop_lost',
    'start_lost',
    'transcript_amplification',
    'inframe_insertion',
    'inframe_deletion',
    'missense_variant',
    'protein_altering_variant',
    'splice_region_variant',
    'incomplete_terminal_codon_variant',
    'start_retained_variant',
    'stop_retained_variant',
    'synonymous_variant',
    'coding_sequence_variant',
    'mature_miRNA_variant',
    '5_prime_UTR_variant',
    '3_prime_UTR_variant',
    'non_coding_transcript_exon_variant',
    'intron_variant',
    'NMD_transcript_variant',
    'non_coding_transcript_variant',
    'upstream_gene_variant',
    'downstream_gene_variant',
    'TFBS_ablation',
    'TFBS_amplification',
    'TF_binding_site_variant',
    'regulatory_region_ablation',
    'regulatory_region_amplification',
    'feature_elongation',
    'regulatory_region_variant',
    'feature_truncation',
    'intergenic_variant'
]

def row_firstann(r):
    """Return the first result in the row.

    Parameters
    ----------
    r : pandas.Series
        VCF row.

    Returns
    -------
    str
        Annotation result. Empty string if annotation is not found.

    See Also
    --------
    row_mostsevere
        Similar method that returns result with the most severe
        consequence in the row.

    Examples
    --------

    >>> data = {
    ...     'CHROM': ['chr1', 'chr1', 'chr2', 'chr2'],
    ...     'POS': [100, 101, 200, 201],
    ...     'ID': ['.', '.', '.', '.'],
    ...     'REF': ['G', 'C', 'G', 'CCCA'],
    ...     'ALT': ['C', 'T', 'A', 'C'],
    ...     'QUAL': ['.', '.', '.', '.'],
    ...     'FILTER': ['.', '.', '.', '.'],
    ...     'INFO': [
    ...         'CSQ=C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000375733.2|protein_coding|||||||||||4617|-1||HGNC|12936||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000375743.4|protein_coding|||||||||||4615|-1||HGNC|12936||||||||||||||||,C|missense_variant|MODERATE|SPEN|ENSG00000065526|Transcript|ENST00000375759.3|protein_coding|12/15||||10322|10118|3373|Q/P|cAg/cCg|||1||HGNC|17575|||||tolerated_low_confidence(0.08)|possibly_damaging(0.718)||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000440560.1|protein_coding|||||||||||4615|-1|cds_start_NF|HGNC|12936||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000462525.1|processed_transcript|||||||||||4615|-1||HGNC|12936||||||||||||||||,C|upstream_gene_variant|MODIFIER|SPEN|ENSG00000065526|Transcript|ENST00000487496.1|processed_transcript|||||||||||1329|1||HGNC|17575||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000537142.1|protein_coding|||||||||||4617|-1||HGNC|12936||||||||||||||||,C|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00000922256|promoter_flanking_region|||||||||||||||||||||||||||||||',
    ...         'CSQ=T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000261443.5|protein_coding||||||||||COSV99795232|3453|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000339438.6|protein_coding||||||||||COSV99795232|3071|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000358528.4|protein_coding||||||||||COSV99795232|3075|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000369530.1|protein_coding||||||||||COSV99795232|3470|-1||HGNC|29905|||||||||1|1||||||,T|missense_variant|MODERATE|NRAS|ENSG00000213281|Transcript|ENST00000369535.4|protein_coding|3/7||||502|248|83|A/D|gCc/gAc|COSV99795232||-1||HGNC|7989|||||deleterious(0.02)|probably_damaging(0.946)|||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000438362.2|protein_coding||||||||||COSV99795232|3074|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000483407.1|processed_transcript||||||||||COSV99795232|3960|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000530886.1|protein_coding||||||||||COSV99795232|3463|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000534699.1|protein_coding||||||||||COSV99795232|4115|-1||HGNC|29905|||||||||1|1||||||',
    ...         'CSQ=A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000358660.3|protein_coding|10/16||||1296|1255|419|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.999)|||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000368196.3|protein_coding|10/16||||1375|1255|419|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|possibly_damaging(0.894)|||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000392302.2|protein_coding|11/17||||1339|1165|389|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.983)|||1|1||||||,A|downstream_gene_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000489021.2|processed_transcript||||||||||COSV62328771|1037|1||HGNC|8031|||||||||1|1||||||,A|stop_gained&NMD_transcript_variant|HIGH|NTRK1|ENSG00000198400|Transcript|ENST00000497019.2|nonsense_mediated_decay|10/16||||1183|1032|344|W/*|tgG/tgA|COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000524377.1|protein_coding|11/17||||1314|1273|425|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.996)|||1|1||||||,A|non_coding_transcript_exon_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000530298.1|retained_intron|12/17||||1313|||||COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|non_coding_transcript_exon_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000534682.1|retained_intron|2/4||||496|||||COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00001787924|open_chromatin_region||||||||||COSV62328771||||||||||||||1|1||||||',
    ...         'CSQ=-|upstream_gene_variant|MODIFIER|LRRC56|ENSG00000161328|Transcript|ENST00000270115.7|protein_coding||||||||||rs1164486792|3230|1||HGNC|25430||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000311189.7|protein_coding|2/6||||198-200|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000397594.1|protein_coding|1/5||||79-81|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000397596.2|protein_coding|2/5||||162-164|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000417302.1|protein_coding|2/6||||214-216|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000451590.1|protein_coding|2/5||||214-216|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000462734.1|processed_transcript||||||||||rs1164486792|700|-1||HGNC|5173||||||||||1||||||,-|non_coding_transcript_exon_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000468682.2|processed_transcript|2/3||||514-516|||||rs1164486792||-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000478324.1|processed_transcript||||||||||rs1164486792|683|-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000479482.1|processed_transcript||||||||||rs1164486792|319|-1||HGNC|5173||||||||||1||||||,-|non_coding_transcript_exon_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000482021.1|processed_transcript|2/2||||149-151|||||rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion&NMD_transcript_variant|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000493230.1|nonsense_mediated_decay|2/7||||202-204|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|downstream_gene_variant|MODIFIER|RP13-46H24.1|ENSG00000254739|Transcript|ENST00000526431.1|antisense||||||||||rs1164486792|4636|1||Clone_based_vega_gene|||||||||||1||||||,-|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00000035647|promoter||||||||||rs1164486792|||||||||||||||1||||||'
    ...     ],
    ...     'FORMAT': ['GT', 'GT', 'GT', 'GT'],
    ...     'Steven': ['0/1', '0/1', '0/1', '0/1'],
    ... }
    >>> vf = pyvcf.VcfFrame.from_dict(meta, data)
    >>> vf.df.apply(pyvep.row_firstann, axis=1)
    0    C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG...
    1    T|downstream_gene_variant|MODIFIER|CSDE1|ENSG0...
    2    A|missense_variant|MODERATE|NTRK1|ENSG00000198...
    3    -|upstream_gene_variant|MODIFIER|LRRC56|ENSG00...
    dtype: object
    """
    results = pyvcf.row_parseinfo(r, 'CSQ')
    if not results:
        return ''
    return results.split(',')[0]

def row_mostsevere(r):
    """Return result with the most severe consequence in the row.

    Parameters
    ----------
    r : pandas.Series
        VCF row.

    Returns
    -------
    str
        Annotation result. Empty string if annotation is not found.

    See Also
    --------
    row_firstann
        Similar method that returns the first result in the row.

    Examples
    --------

    >>> data = {
    ...     'CHROM': ['chr1', 'chr1', 'chr2', 'chr2'],
    ...     'POS': [100, 101, 200, 201],
    ...     'ID': ['.', '.', '.', '.'],
    ...     'REF': ['G', 'C', 'G', 'CCCA'],
    ...     'ALT': ['C', 'T', 'A', 'C'],
    ...     'QUAL': ['.', '.', '.', '.'],
    ...     'FILTER': ['.', '.', '.', '.'],
    ...     'INFO': [
    ...         'CSQ=C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000375733.2|protein_coding|||||||||||4617|-1||HGNC|12936||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000375743.4|protein_coding|||||||||||4615|-1||HGNC|12936||||||||||||||||,C|missense_variant|MODERATE|SPEN|ENSG00000065526|Transcript|ENST00000375759.3|protein_coding|12/15||||10322|10118|3373|Q/P|cAg/cCg|||1||HGNC|17575|||||tolerated_low_confidence(0.08)|possibly_damaging(0.718)||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000440560.1|protein_coding|||||||||||4615|-1|cds_start_NF|HGNC|12936||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000462525.1|processed_transcript|||||||||||4615|-1||HGNC|12936||||||||||||||||,C|upstream_gene_variant|MODIFIER|SPEN|ENSG00000065526|Transcript|ENST00000487496.1|processed_transcript|||||||||||1329|1||HGNC|17575||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000537142.1|protein_coding|||||||||||4617|-1||HGNC|12936||||||||||||||||,C|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00000922256|promoter_flanking_region|||||||||||||||||||||||||||||||',
    ...         'CSQ=T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000261443.5|protein_coding||||||||||COSV99795232|3453|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000339438.6|protein_coding||||||||||COSV99795232|3071|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000358528.4|protein_coding||||||||||COSV99795232|3075|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000369530.1|protein_coding||||||||||COSV99795232|3470|-1||HGNC|29905|||||||||1|1||||||,T|missense_variant|MODERATE|NRAS|ENSG00000213281|Transcript|ENST00000369535.4|protein_coding|3/7||||502|248|83|A/D|gCc/gAc|COSV99795232||-1||HGNC|7989|||||deleterious(0.02)|probably_damaging(0.946)|||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000438362.2|protein_coding||||||||||COSV99795232|3074|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000483407.1|processed_transcript||||||||||COSV99795232|3960|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000530886.1|protein_coding||||||||||COSV99795232|3463|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000534699.1|protein_coding||||||||||COSV99795232|4115|-1||HGNC|29905|||||||||1|1||||||',
    ...         'CSQ=A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000358660.3|protein_coding|10/16||||1296|1255|419|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.999)|||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000368196.3|protein_coding|10/16||||1375|1255|419|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|possibly_damaging(0.894)|||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000392302.2|protein_coding|11/17||||1339|1165|389|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.983)|||1|1||||||,A|downstream_gene_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000489021.2|processed_transcript||||||||||COSV62328771|1037|1||HGNC|8031|||||||||1|1||||||,A|stop_gained&NMD_transcript_variant|HIGH|NTRK1|ENSG00000198400|Transcript|ENST00000497019.2|nonsense_mediated_decay|10/16||||1183|1032|344|W/*|tgG/tgA|COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000524377.1|protein_coding|11/17||||1314|1273|425|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.996)|||1|1||||||,A|non_coding_transcript_exon_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000530298.1|retained_intron|12/17||||1313|||||COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|non_coding_transcript_exon_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000534682.1|retained_intron|2/4||||496|||||COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00001787924|open_chromatin_region||||||||||COSV62328771||||||||||||||1|1||||||',
    ...         'CSQ=-|upstream_gene_variant|MODIFIER|LRRC56|ENSG00000161328|Transcript|ENST00000270115.7|protein_coding||||||||||rs1164486792|3230|1||HGNC|25430||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000311189.7|protein_coding|2/6||||198-200|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000397594.1|protein_coding|1/5||||79-81|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000397596.2|protein_coding|2/5||||162-164|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000417302.1|protein_coding|2/6||||214-216|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000451590.1|protein_coding|2/5||||214-216|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000462734.1|processed_transcript||||||||||rs1164486792|700|-1||HGNC|5173||||||||||1||||||,-|non_coding_transcript_exon_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000468682.2|processed_transcript|2/3||||514-516|||||rs1164486792||-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000478324.1|processed_transcript||||||||||rs1164486792|683|-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000479482.1|processed_transcript||||||||||rs1164486792|319|-1||HGNC|5173||||||||||1||||||,-|non_coding_transcript_exon_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000482021.1|processed_transcript|2/2||||149-151|||||rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion&NMD_transcript_variant|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000493230.1|nonsense_mediated_decay|2/7||||202-204|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|downstream_gene_variant|MODIFIER|RP13-46H24.1|ENSG00000254739|Transcript|ENST00000526431.1|antisense||||||||||rs1164486792|4636|1||Clone_based_vega_gene|||||||||||1||||||,-|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00000035647|promoter||||||||||rs1164486792|||||||||||||||1||||||'
    ...     ],
    ...     'FORMAT': ['GT', 'GT', 'GT', 'GT'],
    ...     'Steven': ['0/1', '0/1', '0/1', '0/1'],
    ... }
    >>> vf = pyvcf.VcfFrame.from_dict(meta, data)
    >>> vf.df.apply(pyvep.row_mostsevere, axis=1)
    0    C|missense_variant|MODERATE|SPEN|ENSG000000655...
    1    T|missense_variant|MODERATE|NRAS|ENSG000002132...
    2    A|stop_gained&NMD_transcript_variant|HIGH|NTRK...
    3    -|inframe_deletion|MODERATE|HRAS|ENSG000001747...
    dtype: object
    """
    results = pyvcf.row_parseinfo(r, 'CSQ')
    if not results:
        return ''
    f = lambda x: SEVERITIY.index(x.split('|')[1].split('&')[0])
    return sorted(results.split(','), key=f)[0]

def parseann(vf, targets, sep=' | ', as_series=False):
    """Parse variant annotations in VcfFrame.

    If there are multiple results per row, the method will use the first
    one for extracting requested data.

    Parameters
    ----------
    vf : VcfFrame
        VcfFrame.
    targets : list
        List of subfield IDs or indicies or both.
    sep : str, default: ' | '
        Separator for joining requested annotations.
    as_series : bool, default: False
        If True, return 1D array instead of VcfFrame.

    Returns
    -------
    VcfFrame or pandas.Series
        Parsed annotations in VcfFrame or 1D array.

    Examples
    --------
    Assume we have the following data:

    >>> meta = [
    ...     '##fileformat=VCFv4.1',
    ...     '##VEP="v104" time="2021-05-20 10:50:12" cache="/net/isilonP/public/ro/ensweb-data/latest/tools/grch37/e104/vep/cache/homo_sapiens/104_GRCh37" db="homo_sapiens_core_104_37@hh-mysql-ens-grch37-web" 1000genomes="phase3" COSMIC="92" ClinVar="202012" HGMD-PUBLIC="20204" assembly="GRCh37.p13" dbSNP="154" gencode="GENCODE 19" genebuild="2011-04" gnomAD="r2.1" polyphen="2.2.2" regbuild="1.0" sift="sift5.2.2"',
    ...     '##INFO=<ID=CSQ,Number=.,Type=String,Description="Consequence annotations from Ensembl VEP. Format: Allele|Consequence|IMPACT|SYMBOL|Gene|Feature_type|Feature|BIOTYPE|EXON|INTRON|HGVSc|HGVSp|cDNA_position|CDS_position|Protein_position|Amino_acids|Codons|Existing_variation|DISTANCE|STRAND|FLAGS|SYMBOL_SOURCE|HGNC_ID|MANE_SELECT|MANE_PLUS_CLINICAL|TSL|APPRIS|SIFT|PolyPhen|AF|CLIN_SIG|SOMATIC|PHENO|PUBMED|MOTIF_NAME|MOTIF_POS|HIGH_INF_POS|MOTIF_SCORE_CHANGE|TRANSCRIPTION_FACTORS">'
    ... ]
    >>> data = {
    ...     'CHROM': ['chr1', 'chr1', 'chr2', 'chr2'],
    ...     'POS': [100, 101, 200, 201],
    ...     'ID': ['.', '.', '.', '.'],
    ...     'REF': ['G', 'C', 'CAG', 'C'],
    ...     'ALT': ['T', 'T', 'C', 'T'],
    ...     'QUAL': ['.', '.', '.', '.'],
    ...     'FILTER': ['.', '.', '.', '.'],
    ...     'INFO': [
    ...         'CSQ=T|missense_variant|MODERATE|MTOR|ENSG00000198793|Transcript|ENST00000361445.4|protein_coding|47/58||||6721|6644|2215|S/Y|tCt/tAt|rs587777894&COSV63868278&COSV63868313||-1||HGNC|3942|||||deleterious(0)|possibly_damaging(0.876)||likely_pathogenic&pathogenic|0&1&1|1&1&1|26619011&27159400&24631838&26018084&27830187|||||',
    ...         'CSQ=T|synonymous_variant|LOW|MTOR|ENSG00000198793|Transcript|ENST00000361445.4|protein_coding|49/58||||6986|6909|2303|L|ctG/ctA|rs11121691&COSV63870864||-1||HGNC|3942|||||||0.2206|benign|0&1|1&1|24996771|||||',
    ...         'CSQ=-|frameshift_variant|HIGH|BRCA2|ENSG00000139618|Transcript|ENST00000380152.3|protein_coding|18/27||||8479-8480|8246-8247|2749|Q/X|cAG/c|rs80359701||1||HGNC|1101||||||||pathogenic||1|26467025&26295337&15340362|||||',
    ...         'CSQ=T|missense_variant|MODERATE|MTOR|ENSG00000198793|Transcript|ENST00000361445.4|protein_coding|30/58||||4516|4439|1480|R/H|cGc/cAc|rs780930764&COSV63868373||-1||HGNC|3942|||||tolerated(0.13)|benign(0)||likely_benign|0&1|1&1||||||'
    ...     ],
    ...     'FORMAT': ['GT', 'GT', 'GT', 'GT'],
    ...     'Steven': ['0/1', '0/1', '0/1', '0/1'],
    ... }
    >>> vf = pyvcf.VcfFrame.from_dict(meta, data)

    We can extract gene symbols:

    >>> pyvep.parseann(vf, ['SYMBOL']).df
      CHROM  POS ID  REF ALT QUAL FILTER   INFO FORMAT Steven
    0  chr1  100  .    G   T    .      .   MTOR     GT    0/1
    1  chr1  101  .    C   T    .      .   MTOR     GT    0/1
    2  chr2  200  .  CAG   C    .      .  BRCA2     GT    0/1
    3  chr2  201  .    C   T    .      .   MTOR     GT    0/1

    If we want to add variant impact:

    >>> pyvep.parseann(vf, ['SYMBOL', 'IMPACT']).df
      CHROM  POS ID  REF ALT QUAL FILTER             INFO FORMAT Steven
    0  chr1  100  .    G   T    .      .  MTOR | MODERATE     GT    0/1
    1  chr1  101  .    C   T    .      .       MTOR | LOW     GT    0/1
    2  chr2  200  .  CAG   C    .      .     BRCA2 | HIGH     GT    0/1
    3  chr2  201  .    C   T    .      .  MTOR | MODERATE     GT    0/1

    If we want to change the delimter:

    >>> pyvep.parseann(vf, ['SYMBOL', 'IMPACT'], sep=',').df
      CHROM  POS ID  REF ALT QUAL FILTER           INFO FORMAT Steven
    0  chr1  100  .    G   T    .      .  MTOR,MODERATE     GT    0/1
    1  chr1  101  .    C   T    .      .       MTOR,LOW     GT    0/1
    2  chr2  200  .  CAG   C    .      .     BRCA2,HIGH     GT    0/1
    3  chr2  201  .    C   T    .      .  MTOR,MODERATE     GT    0/1

    Finally, we can return a 1D array instead of VcfFrame:

    >>> pyvep.parseann(vf, ['SYMBOL'], as_series=True)
    0     MTOR
    1     MTOR
    2    BRCA2
    3     MTOR
    dtype: object
    """
    _targets = [x if isinstance(x, int) else annot_names(vf).index(x) for x in targets]
    def func(r):
        ann = row_firstann(r)
        if not ann:
            return '.'
        ann = ann.split('|')
        ann = sep.join([ann[i] for i in _targets])
        return ann
    s = vf.df.apply(func, axis=1)
    if as_series:
        return s
    new_vf = vf.copy()
    new_vf.df.INFO = s
    return new_vf

def to_frame(vf, as_zero=False):
    """
    Create a DataFrame containing analysis-ready VEP annotation data.

    Parameters
    ----------
    vf : VcfFrame
        VcfFrame object.
    as_zero : bool, default: False
        If True, missing values will be converted to zeros instead of ``NaN``.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing VEP annotation data.
    """
    f = lambda r: pd.Series(row_firstann(r).split('|'))
    df = vf.df.apply(f, axis=1)
    if as_zero:
        df = df.replace('', 0)
    else:
        df = df.replace('', np.nan)
    df.columns = annot_names(vf)
    d = {}
    for col in df.columns:
        if col in DATA_TYPES:
            d[col] = DATA_TYPES[col]
    df = df.astype(d)
    return df

def pick_result(vf, mode='mostsevere'):
    """Return a new VcfFrame after picking one result per row.

    Parameters
    ----------
    vf : VcfFrame
        VcfFrame object.
    mode : {'mostsevere', 'firstann'}, default: 'mostsevere'
        Selection mode:

            - mostsevere: use ``row_mostsevere`` method
            - firstann: use ``row_firstann`` method

    Returns
    -------
    VcfFrame
        New VcfFrame.

    Examples
    --------

    >>> meta = [
    ...     '##fileformat=VCFv4.1',
    ...     '##VEP="v104" time="2021-05-20 10:50:12" cache="/net/isilonP/public/ro/ensweb-data/latest/tools/grch37/e104/vep/cache/homo_sapiens/104_GRCh37" db="homo_sapiens_core_104_37@hh-mysql-ens-grch37-web" 1000genomes="phase3" COSMIC="92" ClinVar="202012" HGMD-PUBLIC="20204" assembly="GRCh37.p13" dbSNP="154" gencode="GENCODE 19" genebuild="2011-04" gnomAD="r2.1" polyphen="2.2.2" regbuild="1.0" sift="sift5.2.2"',
    ...     '##INFO=<ID=CSQ,Number=.,Type=String,Description="Consequence annotations from Ensembl VEP. Format: Allele|Consequence|IMPACT|SYMBOL|Gene|Feature_type|Feature|BIOTYPE|EXON|INTRON|HGVSc|HGVSp|cDNA_position|CDS_position|Protein_position|Amino_acids|Codons|Existing_variation|DISTANCE|STRAND|FLAGS|SYMBOL_SOURCE|HGNC_ID|MANE_SELECT|MANE_PLUS_CLINICAL|TSL|APPRIS|SIFT|PolyPhen|AF|CLIN_SIG|SOMATIC|PHENO|PUBMED|MOTIF_NAME|MOTIF_POS|HIGH_INF_POS|MOTIF_SCORE_CHANGE|TRANSCRIPTION_FACTORS">'
    ... ]
    >>> data = {
    ...     'CHROM': ['chr1', 'chr1', 'chr2', 'chr2'],
    ...     'POS': [100, 101, 200, 201],
    ...     'ID': ['.', '.', '.', '.'],
    ...     'REF': ['G', 'C', 'G', 'CCCA'],
    ...     'ALT': ['C', 'T', 'A', 'C'],
    ...     'QUAL': ['.', '.', '.', '.'],
    ...     'FILTER': ['.', '.', '.', '.'],
    ...     'INFO': [
    ...         'CSQ=C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000375733.2|protein_coding|||||||||||4617|-1||HGNC|12936||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000375743.4|protein_coding|||||||||||4615|-1||HGNC|12936||||||||||||||||,C|missense_variant|MODERATE|SPEN|ENSG00000065526|Transcript|ENST00000375759.3|protein_coding|12/15||||10322|10118|3373|Q/P|cAg/cCg|||1||HGNC|17575|||||tolerated_low_confidence(0.08)|possibly_damaging(0.718)||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000440560.1|protein_coding|||||||||||4615|-1|cds_start_NF|HGNC|12936||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000462525.1|processed_transcript|||||||||||4615|-1||HGNC|12936||||||||||||||||,C|upstream_gene_variant|MODIFIER|SPEN|ENSG00000065526|Transcript|ENST00000487496.1|processed_transcript|||||||||||1329|1||HGNC|17575||||||||||||||||,C|downstream_gene_variant|MODIFIER|ZBTB17|ENSG00000116809|Transcript|ENST00000537142.1|protein_coding|||||||||||4617|-1||HGNC|12936||||||||||||||||,C|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00000922256|promoter_flanking_region|||||||||||||||||||||||||||||||',
    ...         'CSQ=T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000261443.5|protein_coding||||||||||COSV99795232|3453|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000339438.6|protein_coding||||||||||COSV99795232|3071|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000358528.4|protein_coding||||||||||COSV99795232|3075|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000369530.1|protein_coding||||||||||COSV99795232|3470|-1||HGNC|29905|||||||||1|1||||||,T|missense_variant|MODERATE|NRAS|ENSG00000213281|Transcript|ENST00000369535.4|protein_coding|3/7||||502|248|83|A/D|gCc/gAc|COSV99795232||-1||HGNC|7989|||||deleterious(0.02)|probably_damaging(0.946)|||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000438362.2|protein_coding||||||||||COSV99795232|3074|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000483407.1|processed_transcript||||||||||COSV99795232|3960|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000530886.1|protein_coding||||||||||COSV99795232|3463|-1||HGNC|29905|||||||||1|1||||||,T|downstream_gene_variant|MODIFIER|CSDE1|ENSG00000009307|Transcript|ENST00000534699.1|protein_coding||||||||||COSV99795232|4115|-1||HGNC|29905|||||||||1|1||||||',
    ...         'CSQ=A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000358660.3|protein_coding|10/16||||1296|1255|419|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.999)|||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000368196.3|protein_coding|10/16||||1375|1255|419|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|possibly_damaging(0.894)|||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000392302.2|protein_coding|11/17||||1339|1165|389|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.983)|||1|1||||||,A|downstream_gene_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000489021.2|processed_transcript||||||||||COSV62328771|1037|1||HGNC|8031|||||||||1|1||||||,A|stop_gained&NMD_transcript_variant|HIGH|NTRK1|ENSG00000198400|Transcript|ENST00000497019.2|nonsense_mediated_decay|10/16||||1183|1032|344|W/*|tgG/tgA|COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|missense_variant|MODERATE|NTRK1|ENSG00000198400|Transcript|ENST00000524377.1|protein_coding|11/17||||1314|1273|425|A/T|Gcc/Acc|COSV62328771||1||HGNC|8031|||||deleterious(0.01)|probably_damaging(0.996)|||1|1||||||,A|non_coding_transcript_exon_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000530298.1|retained_intron|12/17||||1313|||||COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|non_coding_transcript_exon_variant|MODIFIER|NTRK1|ENSG00000198400|Transcript|ENST00000534682.1|retained_intron|2/4||||496|||||COSV62328771||1||HGNC|8031|||||||||1|1||||||,A|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00001787924|open_chromatin_region||||||||||COSV62328771||||||||||||||1|1||||||',
    ...         'CSQ=-|upstream_gene_variant|MODIFIER|LRRC56|ENSG00000161328|Transcript|ENST00000270115.7|protein_coding||||||||||rs1164486792|3230|1||HGNC|25430||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000311189.7|protein_coding|2/6||||198-200|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000397594.1|protein_coding|1/5||||79-81|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000397596.2|protein_coding|2/5||||162-164|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000417302.1|protein_coding|2/6||||214-216|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000451590.1|protein_coding|2/5||||214-216|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000462734.1|processed_transcript||||||||||rs1164486792|700|-1||HGNC|5173||||||||||1||||||,-|non_coding_transcript_exon_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000468682.2|processed_transcript|2/3||||514-516|||||rs1164486792||-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000478324.1|processed_transcript||||||||||rs1164486792|683|-1||HGNC|5173||||||||||1||||||,-|upstream_gene_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000479482.1|processed_transcript||||||||||rs1164486792|319|-1||HGNC|5173||||||||||1||||||,-|non_coding_transcript_exon_variant|MODIFIER|HRAS|ENSG00000174775|Transcript|ENST00000482021.1|processed_transcript|2/2||||149-151|||||rs1164486792||-1||HGNC|5173||||||||||1||||||,-|inframe_deletion&NMD_transcript_variant|MODERATE|HRAS|ENSG00000174775|Transcript|ENST00000493230.1|nonsense_mediated_decay|2/7||||202-204|26-28|9-10|VG/G|gTGGgc/ggc|rs1164486792||-1||HGNC|5173||||||||||1||||||,-|downstream_gene_variant|MODIFIER|RP13-46H24.1|ENSG00000254739|Transcript|ENST00000526431.1|antisense||||||||||rs1164486792|4636|1||Clone_based_vega_gene|||||||||||1||||||,-|regulatory_region_variant|MODIFIER|||RegulatoryFeature|ENSR00000035647|promoter||||||||||rs1164486792|||||||||||||||1||||||'
    ...     ],
    ...     'FORMAT': ['GT', 'GT', 'GT', 'GT'],
    ...     'Steven': ['0/1', '0/1', '0/1', '0/1'],
    ... }
    >>> vf = pyvcf.VcfFrame.from_dict(meta, data)
    >>> vf1 = pyvep.pick_result(vf, mode='mostsevere')
    >>> vf2 = pyvep.pick_result(vf, mode='firstann')
    >>> vf1.df.INFO
    0    CSQ=C|missense_variant|MODERATE|SPEN|ENSG00000...
    1    CSQ=T|missense_variant|MODERATE|NRAS|ENSG00000...
    2    CSQ=A|stop_gained&NMD_transcript_variant|HIGH|...
    3    CSQ=-|inframe_deletion|MODERATE|HRAS|ENSG00000...
    Name: INFO, dtype: object
    >>> vf2.df.INFO
    0    CSQ=C|downstream_gene_variant|MODIFIER|ZBTB17|...
    1    CSQ=T|downstream_gene_variant|MODIFIER|CSDE1|E...
    2    CSQ=A|missense_variant|MODERATE|NTRK1|ENSG0000...
    3    CSQ=-|upstream_gene_variant|MODIFIER|LRRC56|EN...
    Name: INFO, dtype: object
    """
    funcs = {'mostsevere': row_mostsevere, 'firstann': row_firstann}
    new_vf = vf.copy()
    one_row = lambda r: pyvcf.row_updateinfo(r, 'CSQ', funcs[mode](r))
    new_vf.df.INFO = vf.df.apply(one_row, axis=1)
    return new_vf

def annot_names(vf):
    """Return the list of avaialble consequence annotations in the VcfFrame.

    Parameters
    ----------
    vf : VcfFrame
        VcfFrame.

    Returns
    -------
    list
        List of consequence annotations.
    """
    l = []
    for i, line in enumerate(vf.meta):
        if 'ID=CSQ' in line:
            l = re.search(r'Format: (.*?)">', vf.meta[i]).group(1).split('|')
    return l

def filter_query(vf, expr, opposite=None, as_index=False, as_zero=False):
    """
    Select rows that satisfy the query expression.

    This method essentially wraps the :meth:`pandas.DataFrame.query` method.

    Parameters
    ----------
    expr : str
        The query string to evaluate.
    opposite : bool, default: False
        If True, return rows that don't meet the said criteria.
    as_index : bool, default: False
        If True, return boolean index array instead of VcfFrame.
    as_zero : bool, default: False
        If True, missing values will be converted to zeros instead of ``NaN``.

    Returns
    -------
    VcfFrame or pandas.Series
        Filtered VcfFrame or boolean index array.

    Examples
    --------

    >>> from fuc import common, pyvcf, pyvep
    >>> common.load_dataset('tcga-laml')
    >>> fn = '~/fuc-data/tcga-laml/tcga_laml_vep.vcf'
    >>> vf = pyvcf.VcfFrame.from_file(fn)
    >>> filtered_vf = pyvep.filter_query(vf, "SYMBOL == 'TP53'")
    >>> filtered_vf = pyvep.filter_query(vf, "SYMBOL != 'TP53'")
    >>> filtered_vf = pyvep.filter_query(vf, "SYMBOL != 'TP53'", opposite=True)
    >>> filtered_vf = pyvep.filter_query(vf, "Consequence in ['splice_donor_variant', 'stop_gained']")
    >>> filtered_vf = pyvep.filter_query(vf, "(SYMBOL == 'TP53') and (Consequence.str.contains('stop_gained'))")
    >>> filtered_vf = pyvep.filter_query(vf, "gnomAD_AF < 0.001")
    >>> filtered_vf = pyvep.filter_query(vf, "gnomAD_AF < 0.001", as_zero=True)
    >>> filtered_vf = pyvep.filter_query(vf, "(IMPACT == 'HIGH') or (Consequence.str.contains('missense_variant') and (PolyPhen.str.contains('damaging') or SIFT.str.contains('deleterious')))")
    """
    df = to_frame(vf, as_zero=as_zero)
    i = vf.df.index.isin(df.query(expr).index)
    if opposite:
        i = ~i
    if as_index:
        return i
    return vf.__class__(vf.copy_meta(), vf.df[i])



import sys

from .. import api

description = """
Convert a VCF file to a MAF file.
"""

epilog = f"""
[Example] Convert VCF to MAF:
  $ fuc {api.common._script_name()} in.vcf > out.maf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Convert a VCF file to a MAF file.',
    )
    parser.add_argument('vcf', help='Annotated VCF file.')

def main(args):
    mf = api.pymaf.MafFrame.from_vcf(args.vcf)
    sys.stdout.write(mf.to_string())



"""
The pyfq submodule is designed for working with FASTQ files. It implements
``pyfq.FqFrame`` which stores FASTQ data as ``pandas.DataFrame`` to allow
fast computation and easy manipulation.
"""

import gzip
import pandas as pd

class FqFrame:
    """Class for storing FASTQ data."""
    def __init__(self, df):
        self.df = df

    @property
    def shape(self):
        """int : Number of sequence reads in the FqFrame."""
        return self.df.shape

    def to_file(self, file_path):
        """Write the FqFrame to a FASTQ file."""
        with open(file_path, 'w') as f:
            for line in self.df.stack().to_list():
                f.write(line + '\n')

    def readlen(self):
        """Return a dictionary of read lengths and their counts."""
        return self.df.apply(lambda r: len(r.SEQ), axis=1
            ).value_counts().to_dict()

    @classmethod
    def from_file(cls, fn):
        """Construct FqFrame from a FASTQ file.

        Parameters
        ----------
        fn : str
            FASTQ file path (zipped or unzipped).

        Returns
        -------
        FqFrame
            FqFrame.

        See Also
        --------
        FqFrame
            FqFrame object creation using constructor.
        """
        if fn.endswith('.gz'):
            f = gzip.open(fn, 'rt')
        else:
            f = open(fn)
        data = {'ID': [], 'SEQ': [], 'EXT': [], 'QUAL': []}
        n = 0
        for line in f:
            x = line.strip()
            if n == 0:
                data['ID'].append(x)
                n += 1
            elif n == 1:
                data['SEQ'].append(x)
                n += 1
            elif n == 2:
                data['EXT'].append(x)
                n += 1
            else:
                data['QUAL'].append(x)
                n = 0
        df = pd.DataFrame.from_dict(data)
        f.close()
        return cls(df)



import sys

from .. import api

description = """
Count allelic depth from a SAM/BAM/CRAM file.
"""

epilog = f"""
[Example] Provide sites with a TSV file:
  $ fuc {api.common._script_name()} in.bam sites.tsv > out.tsv

[Example] Provide sites with a VCF file:
  $ fuc {api.common._script_name()} in.bam sites.vcf > out.tsv
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Compute allelic depth from a SAM/BAM/CRAM file.',
    )
    parser.add_argument(
        'bam',
        help='Alignment file.'
    )
    parser.add_argument(
        'sites',
        help='TSV file containing two columns, chromosome and position. \n'
             'This can also be a BED or VCF file (zipped or unzipped). \n'
             'Input type will be detected automatically.'
    )

def main(args):
    if '.vcf' in args.sites:
        vf = api.pyvcf.VcfFrame.from_file(args.sites)
        def one_row(r):
            if api.pyvcf.row_hasindel(r):
                return f'{r.CHROM}-{r.POS+1}'
            else:
                return f'{r.CHROM}-{r.POS}'
        sites = vf.df.apply(one_row, axis=1).to_list()
    elif '.bed' in args.sites:
        bf = api.pybed.BedFrame.from_file(args.sites)
        sites = bf.gr.df.apply(lambda r: f'{r.Chromosome}-{r.Start}', axis=1).to_list()
    else:
        with open(args.sites) as f:
            sites = []
            for line in f:
                fields = line.strip().split('\t')
                sites.append(fields[0] + '-' + fields[1])
    df = api.pybam.count_allelic_depth(args.bam, sites)
    sys.stdout.write(df.to_csv(sep='\t', index=False))



"""
The pycov submodule is designed for working with depth of coverage data
from sequence alingment files (SAM/BAM/CRAM). It implements
``pycov.CovFrame`` which stores read depth data as ``pandas.DataFrame`` via
the `pysam <https://pysam.readthedocs.io/en/latest/api.html>`_ package to
allow fast computation and easy manipulation. The ``pycov.CovFrame`` class
also contains many useful plotting methods such as ``CovFrame.plot_region``
and ``CovFrame.plot_uniformity``.
"""
from io import StringIO, IOBase
import gzip

from . import common, pybam, pybed

import pysam
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

def concat(cfs, axis=0):
    """
    Concatenate CovFrame objects along a particular axis.

    Parameters
    ----------
    cfs : list
        List of CovFrame objects.
    axis : {0/'index', 1/'columns'}, default: 0
        The axis to concatenate along.

    Returns
    -------
    CovFrame
        Concatenated CovFrame.
    """
    if axis:
        df = pd.concat([x.df.set_index(['Chromosome', 'Position'])
            for x in cfs], axis=axis).reset_index()
    else:
        df = pd.concat([x.df for x in cfs], axis=axis)
    return CovFrame(df)

def simulate(mode='wgs', loc=30, scale=5, size=1000):
    """
    Simulate read depth data for single sample.

    Generated read depth will be integer and non-negative.

    Parameters
    ----------
    mode : {'wgs'}, default: 'wgs'
        Additional modes will be made available in future releases.
    loc : float, default: 30
        Mean ("centre") of the distribution.
    scale : float, default: 5
        Standard deviation (spread or "width") of the distribution. Must be
        non-negative.
    size : int, default: 1000
        Number of base pairs to return.

    Returns
    -------
    numpy.ndarray
        Numpy array object.

    Examples
    --------

    >>> from fuc import pycov
    >>> pycov.simulate(size=10)
    array([25, 32, 30, 31, 26, 25, 33, 29, 28, 35])
    """
    a = np.random.normal(loc=loc, scale=scale, size=size)

    # Read depth must be integer.
    a = a.astype(int)

    # Read dpeth must be non-negative.
    a = np.absolute(a)

    return a

class CovFrame:
    """
    Class for storing read depth data from one or more SAM/BAM/CRAM files.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing read depth data.

    See Also
    --------
    CovFrame.from_bam
        Construct CovFrame from one or more SAM/BAM/CRAM files.
    CovFrame.from_dict
        Construct CovFrame from dict of array-like or dicts.
    CovFrame.from_file
        Construct CovFrame from a text file containing read depth data.

    Examples
    --------

    >>> import numpy as np
    >>> import pandas as pd
    >>> from fuc import pycov
    >>> data = {
    ...     'Chromosome': ['chr1'] * 1000,
    ...     'Position': np.arange(1000, 2000),
    ...     'A': pycov.simulate(loc=35, scale=5),
    ...     'B': pycov.simulate(loc=25, scale=7),
    ... }
    >>> df = pd.DataFrame(data)
    >>> cf = pycov.CovFrame(df)
    >>> cf.df.head()
      Chromosome  Position   A   B
    0       chr1      1000  22  23
    1       chr1      1001  34  30
    2       chr1      1002  33  27
    3       chr1      1003  32  21
    4       chr1      1004  32  15
    """
    def _check_df(self, df):
        df = df.reset_index(drop=True)
        df.Chromosome = df.Chromosome.astype(str)
        return df

    def __init__(self, df):
        self._df = self._check_df(df)

    @property
    def df(self):
        """pandas.DataFrame : DataFrame containing read depth data."""
        return self._df

    @df.setter
    def df(self, value):
        self._df = self._check_df(value)

    @property
    def contigs(self):
        """list : List of contig names."""
        return list(self.df.Chromosome.unique())

    @property
    def has_chr_prefix(self):
        """bool : Whether the (annoying) 'chr' string is found."""
        for contig in self.contigs:
            if 'chr' in contig:
                return True
        return False

    @property
    def samples(self):
        """list : List of the sample names."""
        return self.df.columns[2:].to_list()

    @property
    def shape(self):
        """tuple : Dimensionality of CovFrame (positions, samples)."""
        return (self.df.shape[0], len(self.samples))

    @classmethod
    def from_bam(
        cls, bam=None, fn=None, bed=None, region=None, zero=False,
        map_qual=None, names=None
    ):
        """
        Construct CovFrame from one or more SAM/BAM/CRAM files.

        Alignment files must be specified with either ``bam`` or ``fn``, but
        it's an error to use both.

        By default, the method will count all reads within the alignment
        files. You can specify target regions with either ``bed`` or
        ``region``, but not both. When you do this, pay close attention to
        the 'chr' string in contig names (e.g. 'chr1' vs. '1'). Note also
        that ``region`` requires the input files be indexed.

        Under the hood, the method computes read depth using the
        :command:`samtools depth` command.

        Parameters
        ----------
        bam : str or list, optional
            One or more alignment files.
        fn : str, optional
            File containing one alignment file per line.
        bed : str, optional
            BED file.
        region : str, optional
            Target region ('chrom:start-end').
        zero : bool, default: False
            If True, output all positions (including those with zero depth).
        map_qual: int, optional
            Only count reads with mapping quality greater than orequal to
            this number.
        names : list, optional
            Use these as sample names instead of the SM tags.

        Returns
        -------
        CovFrame
            CovFrame object.

        See Also
        --------
        CovFrame
            CovFrame object creation using constructor.
        CovFrame.from_dict
            Construct CovFrame from dict of array-like or dicts.
        CovFrame.from_file
            Construct CovFrame from a text file containing read depth data.

        Examples
        --------

        >>> from fuc import pycov
        >>> cf = pycov.CovFrame.from_bam(bam)
        >>> cf = pycov.CovFrame.from_bam([bam1, bam2])
        >>> cf = pycov.CovFrame.from_bam(bam, region='19:41497204-41524301')
        """
        bam_files = []

        if bam is None and fn is None:
            raise ValueError(
                "Either the 'bam' or 'fn' parameter must be provided.")
        elif bam is not None and fn is not None:
            raise ValueError(
                "The 'bam' and 'fn' parameters cannot be used together.")
        elif bam is not None and fn is None:
            if isinstance(bam, str):
                bam_files.append(bam)
            else:
                bam_files += bam
        else:
            bam_files += common.convert_file2list(fn)

        args = []

        if zero:
            args += ['-a']
        if region is not None:
            args += ['-r', region]
        if map_qual is not None:
            args += ['-Q', str(map_qual)]
        if bed is not None:
            args += ['-b', bed]

        args += bam_files
        s = pysam.depth(*args)
        headers = ['Chromosome', 'Position']
        dtype = {'Chromosome': str,'Position': int}
        for i, bam_file in enumerate(bam_files):
            if names:
                name = names[i]
            else:
                samples = pybam.tag_sm(bam_file)
                if len(samples) > 1:
                    m = f'multiple sample names detected: {bam_file}'
                    raise ValueError(m)
                name = samples[0]
            headers.append(name)
            dtype[name] = int
        df = pd.read_csv(StringIO(s), sep='\t', header=None,
                         names=headers, dtype=dtype)
        return cls(df)

    @classmethod
    def from_dict(cls, data):
        """
        Construct CovFrame from dict of array-like or dicts.

        Parameters
        ----------
        data : dict
            Of the form {field : array-like} or {field : dict}.

        Returns
        -------
        CovFrame
            CovFrame object.

        See Also
        --------
        CovFrame
            CovFrame object creation using constructor.
        CovFrame.from_bam
            Construct CovFrame from one or more SAM/BAM/CRAM files.
        CovFrame.from_file
            Construct CovFrame from a text file containing read depth data.

        Examples
        --------

        >>> import numpy as np
        >>> from fuc import pycov
        >>> data = {
        ...     'Chromosome': ['chr1'] * 1000,
        ...     'Position': np.arange(1000, 2000),
        ...     'A': pycov.simulate(loc=35, scale=5),
        ...     'B': pycov.simulate(loc=25, scale=7),
        ... }
        >>> cf = pycov.CovFrame.from_dict(data)
        >>> cf.df.head()
          Chromosome  Position   A   B
        0       chr1      1000  36  22
        1       chr1      1001  39  35
        2       chr1      1002  33  19
        3       chr1      1003  36  20
        4       chr1      1004  31  24
        """
        return cls(pd.DataFrame(data))

    @classmethod
    def from_file(cls, fn, compression=False):
        """
        Construct CovFrame from a TSV file containing read depth data.

        Parameters
        ----------
        fn : str or file-like object
            TSV file (zipped or unzipped). By file-like object, we refer to
            objects with a :meth:`read()` method, such as a file handle.
        compression : bool, default: False
            If True, use GZIP decompression regardless of filename.

        Returns
        -------
        CovFrame
            CovFrame object.

        See Also
        --------
        CovFrame
            CovFrame object creation using constructor.
        CovFrame.from_bam
            Construct CovFrame from one or more SAM/BAM/CRAM files.
        CovFrame.from_dict
            Construct CovFrame from dict of array-like or dicts.

        Examples
        --------

        >>> from fuc import pycov
        >>> cf = pycov.CovFrame.from_file('unzipped.tsv')
        >>> cf = pycov.CovFrame.from_file('zipped.tsv.gz')
        >>> cf = pycov.CovFrame.from_file('zipped.tsv', compression=True)
        """
        if isinstance(fn, IOBase):
            return cls(pd.read_table(fn))

        if fn.endswith('.gz') or compression:
            f = gzip.open(fn, 'rt')
        else:
            f = open(fn)

        headers = f.readline().strip().split('\t')

        f.close()

        if 'Chromosome' not in headers:
            raise ValueError("Input file is missing 'Chromosome' column")

        if 'Position' not in headers:
            raise ValueError("Input file is missing 'Position' column")

        dtype = {}

        for header in headers:
            if header == 'Chromosome':
                dtype[header] = str
            elif header == 'Position':
                dtype[header] = float
            else:
                dtype[header] = float

        return cls(pd.read_table(fn, dtype=dtype))

    def plot_region(
        self, sample, region=None, samples=None, label=None, ax=None, figsize=None,
        **kwargs
    ):
        """
        Create read depth profile for specified region.

        Region can be omitted if there is only one contig in the CovFrame.

        Parameters
        ----------
        region : str, optional
            Target region ('chrom:start-end').
        label : str, optional
            Label to use for the data points.
        ax : matplotlib.axes.Axes, optional
            Pre-existing axes for the plot. Otherwise, crete a new one.
        figsize : tuple, optional
            Width, height in inches. Format: (float, float).
        kwargs
            Other keyword arguments will be passed down to
            :meth:`seaborn.lineplot`.

        Returns
        -------
        matplotlib.axes.Axes
            The matplotlib axes containing the plot.

        Examples
        --------
        Below is a simple example:

        .. plot::
            :context: close-figs

            >>> import matplotlib.pyplot as plt
            >>> import numpy as np
            >>> from fuc import pycov
            >>> data = {
            ...    'Chromosome': ['chr1'] * 1000,
            ...    'Position': np.arange(1000, 2000),
            ...     'A': pycov.simulate(loc=35, scale=5),
            ...     'B': pycov.simulate(loc=25, scale=7),
            ... }
            >>> cf = pycov.CovFrame.from_dict(data)
            >>> ax = cf.plot_region('A')
            >>> plt.tight_layout()

        We can draw multiple profiles in one plot:

        .. plot::
            :context: close-figs

            >>> ax = cf.plot_region('A', label='A')
            >>> cf.plot_region('B', label='B', ax=ax)
            >>> ax.legend()
            >>> plt.tight_layout()
        """
        if region is None:
            if len(self.contigs) == 1:
                cf = self.copy()
            else:
                raise ValueError('Multiple contigs found')
        else:
            cf = self.slice(region)

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        sns.lineplot(
            data=cf.df, x='Position', y=sample, ax=ax, label=label, **kwargs
        )

        ax.set_ylabel('Depth')

        return ax

    def slice(self, region):
        """
        Slice the CovFrame for the region.

        Parameters
        ----------
        region : str
            Region (‘chrom:start-end’).

        Returns
        -------
        CovFrame
            Sliced CovFrame.

        Examples
        --------

        >>> import numpy as np
        >>> from fuc import pycov
        >>> data = {
        ...     'Chromosome': ['chr1']*500 + ['chr2']*500,
        ...     'Position': np.arange(1000, 2000),
        ...     'A': pycov.simulate(loc=35, scale=5),
        ...     'B': pycov.simulate(loc=25, scale=7),
        ... }
        >>> cf = pycov.CovFrame.from_dict(data)
        >>> cf.slice('chr2').df.head()
          Chromosome  Position   A   B
        0       chr2      1500  37  34
        1       chr2      1501  28  12
        2       chr2      1502  35  29
        3       chr2      1503  34  34
        4       chr2      1504  32  21
        >>> cf.slice('chr2:1500-1504').df
          Chromosome  Position   A   B
        0       chr2      1500  37  34
        1       chr2      1501  28  12
        2       chr2      1502  35  29
        3       chr2      1503  34  34
        4       chr2      1504  32  21
        >>> cf.slice('chr2:-1504').df
          Chromosome  Position   A   B
        0       chr2      1500  37  34
        1       chr2      1501  28  12
        2       chr2      1502  35  29
        3       chr2      1503  34  34
        4       chr2      1504  32  21
        """
        chrom, start, end = common.parse_region(region)
        df = self.df[self.df.Chromosome == chrom]
        if not pd.isna(start):
            df = df[df.Position >= start]
        if not pd.isna(end):
            df = df[df.Position <= end]
        return self.__class__(df)

    def to_string(self):
        """
        Render the CovFrame to a console-friendly tabular output.

        Returns
        -------
        str
            String representation of the CovFrame.
        """
        return self.df.to_csv(index=False, sep='\t')

    def plot_uniformity(
        self, mode='aggregated', frac=0.1, n=20, m=None, marker=None,
        ax=None, figsize=None, **kwargs
    ):
        """
        Create a line plot visualizing the uniformity in read depth.

        Parameters
        ----------
        mode : {'aggregated', 'individual'}, default: 'aggregated'
            Determines how to display the lines:

            - 'aggregated': Aggregate over repeated values to show the mean
              and 95% confidence interval.
            - 'individual': Show data for individual samples.

        frac : float, default: 0.1
            Fraction of data to be sampled (to speed up the process).
        n : int or list, default: 20
            Number of evenly spaced points to generate for the x-axis.
            Alternatively, positions can be manually specified by providing
            a list.
        m : float, optional
            Maximum point in the x-axis. By default, it will be the maximum
            depth in the entire dataset.
        marker : str, optional
            Marker style string (e.g. 'o').
        ax : matplotlib.axes.Axes, optional
            Pre-existing axes for the plot. Otherwise, crete a new one.
        figsize : tuple, optional
            Width, height in inches. Format: (float, float).
        kwargs
            Other keyword arguments will be passed down to
            :meth:`seaborn.lineplot`.

        Returns
        -------
        matplotlib.axes.Axes
            The matplotlib axes containing the plot.

        Examples
        --------
        By default (``mode='aggregated'``), the method will aggregate over
        repeated values:

        .. plot::
            :context: close-figs

            >>> import matplotlib.pyplot as plt
            >>> import numpy as np
            >>> from fuc import pycov
            >>> data = {
            ...     'Chromosome': ['chr1'] * 1000,
            ...     'Position': np.arange(1000, 2000),
            ...     'A': pycov.simulate(loc=35, scale=5),
            ...     'B': pycov.simulate(loc=25, scale=7),
            ... }
            >>> cf = pycov.CovFrame.from_dict(data)
            >>> cf.plot_uniformity(mode='aggregated')
            >>> plt.tight_layout()

        We can display data for individual samples:

        .. plot::
            :context: close-figs

            >>> cf.plot_uniformity(mode='individual')
            >>> plt.tight_layout()
        """
        # Sample positions to speed up the process.
        df = self.df.sample(frac=frac)

        # Determine x-axis points.
        if isinstance(n, list):
            coverages = n
        else:
            if m is None:
                m = df.iloc[:, 2:].max().max()
            coverages = np.linspace(1, m, n, endpoint=True)

        data = {'Coverage': coverages}

        for sample in self.samples:
            fractions = []
            for coverage in coverages:
                count = sum(df[sample] >= coverage)
                fractions.append(count / df.shape[0])
            data[sample] = fractions
        df = pd.DataFrame(data)
        df = df.melt(id_vars=['Coverage'], var_name='Sample',
            value_name='Fraction')

        # Determines how to display the lines.
        if mode == 'aggregated':
            hue = None
        else:
            hue = 'Sample'

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        sns.lineplot(
            x='Coverage', y='Fraction', data=df, hue=hue, marker=marker,
            ax=ax, **kwargs
        )

        ax.set_xlabel('Coverage')
        ax.set_ylabel('Fraction of sampled bases >= coverage')

        return ax

    def plot_distribution(
        self, mode='aggregated', frac=0.1, ax=None, figsize=None, **kwargs
    ):
        """
        Create a line plot visualizaing the distribution of per-base read depth.

        Parameters
        ----------
        mode : {'aggregated', 'individual'}, default: 'aggregated'
            Determines how to display the lines:

            - 'aggregated': Aggregate over repeated values to show the mean
              and 95% confidence interval.
            - 'individual': Show data for individual samples.

        frac : float, default: 0.1
            Fraction of data to be sampled (to speed up the process).
        ax : matplotlib.axes.Axes, optional
            Pre-existing axes for the plot. Otherwise, crete a new one.
        figsize : tuple, optional
            Width, height in inches. Format: (float, float).
        kwargs
            Other keyword arguments will be passed down to
            :meth:`seaborn.lineplot`.

        Returns
        -------
        matplotlib.axes.Axes
            The matplotlib axes containing the plot.

        Examples
        --------
        By default (``mode='aggregated'``), the method will aggregate over
        repeated values:

        .. plot::
            :context: close-figs

            >>> import matplotlib.pyplot as plt
            >>> import numpy as np
            >>> from fuc import pycov
            >>> data = {
            ...     'Chromosome': ['chr1'] * 1000,
            ...     'Position': np.arange(1000, 2000),
            ...     'A': pycov.simulate(loc=35, scale=5),
            ...     'B': pycov.simulate(loc=25, scale=7),
            ... }
            >>> cf = pycov.CovFrame.from_dict(data)
            >>> cf.plot_distribution(mode='aggregated', frac=0.9)
            >>> plt.tight_layout()

        We can display data for individual samples:

        .. plot::
            :context: close-figs

            >>> cf.plot_distribution(mode='individual', frac=0.9)
            >>> plt.tight_layout()
        """
        # Sample positions to speed up the process.
        df = self.df.sample(frac=frac)

        def one_col(c):
            s = c.value_counts()
            s = s / s.sum()
            return s

        df = df.iloc[:, 2:].apply(one_col)
        df = df.fillna(0)
        df = df.reset_index().rename(columns=dict(index='Coverage'))
        df = df.melt(id_vars='Coverage', var_name='Sample',
            value_name='Fraction')

        # Determines how to display the lines.
        if mode == 'aggregated':
            hue = None
        else:
            hue = 'Sample'

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        sns.lineplot(
            x='Coverage', y='Fraction', data=df, hue=hue, ax=ax, **kwargs
        )

        ax.set_xlabel('Coverage')
        ax.set_ylabel('Fraction of sampled bases')

        return ax

    def copy_df(self):
        """Return a copy of the dataframe."""
        return self.df.copy()

    def copy(self):
        """Return a copy of the CovFrame."""
        return self.__class__(self.copy_df())

    def to_string(self):
        """
        Render the CovFrame to a console-friendly tabular output.

        Returns
        -------
        str
            String representation of the CovFrame.
        """
        return self.df.to_csv(index=False, sep='\t')

    def to_file(self, fn, compression=False):
        """
        Write the CovFrame to a TSV file.

        If the file name ends with '.gz', the method will automatically
        use the GZIP compression when writing the file.

        Parameters
        ----------
        fn : str
            TSV file (zipped or unzipped).
        compression : bool, default: False
            If True, use the GZIP compression.
        """
        if fn.endswith('.gz') or compression:
            f = gzip.open(fn, 'wt')
        else:
            f = open(fn, 'w')
        f.write(self.to_string())
        f.close()

    def mask_bed(self, bed, opposite=False):
        """
        Mask rows that overlap with BED data.

        Parameters
        ----------
        bed : pybed.BedFrame or str
            BedFrame object or BED file.
        opposite : bool, default: False
            If True, mask rows that don't overlap with BED data.

        Returns
        -------
        CovFrame
            Masked CovFrame.

        Examples
        --------
        Assume we have the following data:

        >>> import numpy as np
        >>> from fuc import pycov, pybed
        >>> data = {
        ...     'Chromosome': ['chr1'] * 1000,
        ...     'Position': np.arange(1000, 2000),
        ...     'A': pycov.simulate(loc=35, scale=5),
        ...     'B': pycov.simulate(loc=25, scale=7),
        ... }
        >>> cf = pycov.CovFrame.from_dict(data)
        >>> cf.df.head()
          Chromosome  Position   A   B
        0       chr1      1000  34  31
        1       chr1      1001  31  20
        2       chr1      1002  41  22
        3       chr1      1003  28  41
        4       chr1      1004  34  23
        >>> data = {
        ...     'Chromosome': ['chr1', 'chr1'],
        ...     'Start': [1000, 1003],
        ...     'End': [1002, 1004]
        ... }
        >>> bf = pybed.BedFrame.from_dict([], data)
        >>> bf.gr.df
          Chromosome  Start   End
        0       chr1   1000  1002
        1       chr1   1003  1004

        We can mask rows that overlap with the BED data:

        >>> cf.mask_bed(bf).df.head()
          Chromosome  Position     A     B
        0       chr1      1000   NaN   NaN
        1       chr1      1001   NaN   NaN
        2       chr1      1002  41.0  22.0
        3       chr1      1003   NaN   NaN
        4       chr1      1004  34.0  23.0

        We can also do the opposite:

        >>> cf.mask_bed(bf, opposite=True).df.head()
          Chromosome  Position     A     B
        0       chr1      1000  34.0  31.0
        1       chr1      1001  31.0  20.0
        2       chr1      1002   NaN   NaN
        3       chr1      1003  28.0  41.0
        4       chr1      1004   NaN   NaN
        """
        if isinstance(bed, pybed.BedFrame):
            bf = bed
        else:
            bf = pybed.BedFrame.from_file(bed)
        def one_row(r):
            if opposite:
                if bf.gr[r.Chromosome, r.Position:r.Position+1].empty:
                    r[2:] = r[2:].apply(lambda x: np.nan)
            else:
                if not bf.gr[r.Chromosome, r.Position:r.Position+1].empty:
                    r[2:] = r[2:].apply(lambda x: np.nan)
            return r
        df = self.df.apply(one_row, axis=1)
        return self.__class__(df)

    def update_chr_prefix(self, mode='remove'):
        """
        Add or remove the (annoying) 'chr' string from the Chromosome column.

        Parameters
        ----------
        mode : {'add', 'remove'}, default: 'remove'
            Whether to add or remove the 'chr' string.

        Returns
        -------
        CovFrame
            Updated CovFrame.

        Examples
        --------

        >>> import numpy as np
        >>> from fuc import pycov
        >>> data = {
        ...     'Chromosome': ['chr1'] * 3 + ['2'] * 3,
        ...     'Position': np.arange(1, 7),
        ...     'A': pycov.simulate(loc=35, scale=5, size=6),
        ...     'B': pycov.simulate(loc=25, scale=7, size=6),
        ... }
        >>> cf = pycov.CovFrame.from_dict(data)
        >>> cf.df
          Chromosome  Position   A   B
        0       chr1         1  35  25
        1       chr1         2  23  14
        2       chr1         3  32  23
        3          2         4  38  25
        4          2         5  33   8
        5          2         6  21  22
        >>> cf.update_chr_prefix(mode='remove').df
          Chromosome  Position   A   B
        0          1         1  35  25
        1          1         2  23  14
        2          1         3  32  23
        3          2         4  38  25
        4          2         5  33   8
        5          2         6  21  22
        >>> cf.update_chr_prefix(mode='add').df
          Chromosome  Position   A   B
        0       chr1         1  35  25
        1       chr1         2  23  14
        2       chr1         3  32  23
        3       chr2         4  38  25
        4       chr2         5  33   8
        5       chr2         6  21  22
        """
        if mode == 'remove':
            def one_row(r):
                r.Chromosome = r.Chromosome.replace('chr', '')
                return r
        elif mode == 'add':
            def one_row(r):
                if 'chr' not in r.Chromosome:
                    r.Chromosome = 'chr' + r.Chromosome
                return r
        else:
            raise ValueError(f'Incorrect mode: {mode}')
        df = self.df.apply(one_row, axis=1)
        return self.__class__(df)

    def subset(self, samples, exclude=False):
        """
        Subset CovFrame for specified samples.

        Parameters
        ----------
        samples : str or list
            Sample name or list of names (the order matters).
        exclude : bool, default: False
            If True, exclude specified samples.

        Returns
        -------
        CovFrame
            Subsetted CovFrame.

        Examples
        --------
        Assume we have the following data:

        >>> import numpy as np
        >>> from fuc import pycov
        >>> data = {
        ...     'Chromosome': ['chr1'] * 1000,
        ...     'Position': np.arange(1000, 2000),
        ...     'A': pycov.simulate(loc=35, scale=5),
        ...     'B': pycov.simulate(loc=25, scale=7),
        ...     'C': pycov.simulate(loc=15, scale=2),
        ...     'D': pycov.simulate(loc=45, scale=8),
        ... }
        >>> cf = pycov.CovFrame.from_dict(data)
        >>> cf.df.head()
          Chromosome  Position   A   B   C   D
        0       chr1      1000  30  30  15  37
        1       chr1      1001  25  24  11  43
        2       chr1      1002  33  24  16  50
        3       chr1      1003  29  22  15  46
        4       chr1      1004  34  30  11  32

        We can subset the CovFrame for the samples A and B:

        >>> cf.subset(['A', 'B']).df.head()
          Chromosome  Position   A   B
        0       chr1      1000  30  30
        1       chr1      1001  25  24
        2       chr1      1002  33  24
        3       chr1      1003  29  22
        4       chr1      1004  34  30

        Alternatively, we can exclude those samples:

        >>> cf.subset(['A', 'B'], exclude=True).df.head()
          Chromosome  Position   C   D
        0       chr1      1000  15  37
        1       chr1      1001  11  43
        2       chr1      1002  16  50
        3       chr1      1003  15  46
        4       chr1      1004  11  32
        """
        if isinstance(samples, str):
            samples = [samples]
        if exclude:
            samples = [x for x in self.samples if x not in samples]
        cols = self.df.columns[:2].to_list() + samples
        return self.__class__(self.df[cols])

    def rename(self, names, indicies=None):
        """
        Rename the samples.

        Parameters
        ----------
        names : dict or list
            Dict of old names to new names or list of new names.
        indicies : list or tuple, optional
            List of 0-based sample indicies. Alternatively, a tuple
            (int, int) can be used to specify an index range.

        Returns
        -------
        CovFrame
            Updated CovFrame.

        Examples
        --------

        >>> import numpy as np
        >>> from fuc import pycov
        >>> data = {
        ...     'Chromosome': ['chr1'] * 2,
        ...     'Position': np.arange(1, 3),
        ...     'A': pycov.simulate(loc=35, scale=5, size=2),
        ...     'B': pycov.simulate(loc=25, scale=7, size=2),
        ...     'C': pycov.simulate(loc=25, scale=7, size=2),
        ...     'D': pycov.simulate(loc=25, scale=7, size=2),
        ... }
        >>> cf = pycov.CovFrame.from_dict(data)
        >>> cf.df
          Chromosome  Position   A   B   C   D
        0       chr1         1  31  19  28  15
        1       chr1         2  35  24  22  17
        >>> cf.rename(['1', '2', '3', '4']).df
          Chromosome  Position   1   2   3   4
        0       chr1         1  31  19  28  15
        1       chr1         2  35  24  22  17
        >>> cf.rename({'B': '2', 'C': '3'}).df
          Chromosome  Position   A   2   3   D
        0       chr1         1  31  19  28  15
        1       chr1         2  35  24  22  17
        >>> cf.rename(['2', '4'], indicies=[1, 3]).df
          Chromosome  Position   A   2   C   4
        0       chr1         1  31  19  28  15
        1       chr1         2  35  24  22  17
        >>> cf.rename(['2', '3'], indicies=(1, 3)).df
          Chromosome  Position   A   2   3   D
        0       chr1         1  31  19  28  15
        1       chr1         2  35  24  22  17
        """
        samples = common.rename(self.samples, names, indicies=indicies)
        columns = self.df.columns[:2].to_list() + samples
        cf = self.copy()
        cf.df.columns = columns
        return cf



import os
from pathlib import Path

from .. import api

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

description = """
Parse the Reports directory from bcl2fastq.

This command will parse, and extract various statistics from, HTML files in
the Reports directory created by the bcl2fastq or bcl2fastq2 prograrm. After
creating an output directory, the command will write the following files:
  - flowcell-summary.csv
  - lane-summary.csv
  - top-unknown-barcodes.csv
  - reports.pdf

Use --sheet to sort samples in the lane-summary.csv file in the same order
as your SampleSheet.csv file. You can also provide a modified version of your
SampleSheet.csv file to subset samples for the lane-summary.csv and
reports.pdf files.
"""

epilog = f"""
[Example] Parse a Reports directory:
  $ fuc {api.common._script_name()} Reports output

[Example] Sort and/or subset samples:
  $ fuc {api.common._script_name()} Reports output --sheet SampleSheet.csv
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        help='Parse the Reports directory from bcl2fastq.',
        description=description,
    )
    parser.add_argument(
        'reports',
        help='Reports directory.'
    )
    parser.add_argument(
        'output',
        help='Output directory (will be created).'
    )
    parser.add_argument(
        '--sheet',
        metavar='PATH',
        help='SampleSheet.csv file. Used for sorting and/or subsetting \n'
             'samples.'
    )

def main(args):
    for path in Path(args.reports).rglob('all/all/all/laneBarcode.html'):
        html_file = path.absolute()

    dfs = pd.read_html(html_file)

    d = {
        '% of thelane': '% of the lane',
        '% Perfectbarcode': '% Perfect barcode',
        '% One mismatchbarcode': '% One mismatch barcode',
        '% PFClusters': '% PF Clusters',
        '% >= Q30bases': '% >= Q30 bases',
        'Mean QualityScore': 'Mean Quality Score',
    }

    df1 = dfs[1]
    df2 = dfs[2].rename(columns=d)
    df2.Sample = df2.Sample.astype(str)
    df3 = dfs[3].dropna()

    if args.sheet is None:
        df4 = pd.DataFrame()
    else:
        with open(args.sheet) as f:
            for i, line in enumerate(f):
                if line.startswith('[Data]'):
                    n = i + 1
        df4 = pd.read_csv(args.sheet, skiprows=n)
        df4.Sample_Name = df4.Sample_Name.astype(str)
        sample_order = df4.Sample_Name.to_list() + ['Undetermined']
        df2.index = df2.Sample
        df2 = df2.loc[sample_order]
        df2 = df2.reset_index(drop=True)

    fig, axes = plt.subplots(3, 2, figsize=(14, 17))

    kwargs = dict(data=df2[:-1], kde=True)

    sns.histplot(ax=axes[0][0], x='PF Clusters', **kwargs)
    sns.histplot(ax=axes[0][1], x='% of the lane', **kwargs)
    sns.histplot(ax=axes[1][0], x='Yield (Mbases)', **kwargs)
    sns.histplot(ax=axes[1][1], x='% PF Clusters', **kwargs)
    sns.histplot(ax=axes[2][0], x='% >= Q30 bases', **kwargs)
    sns.histplot(ax=axes[2][1], x='Mean Quality Score', **kwargs)

    os.mkdir(args.output)

    plt.tight_layout()
    plt.savefig(f'{args.output}/reports.pdf')

    df1_string = df1.to_csv(index=False)
    df2_string = df2.to_csv(index=False, na_rep='NaN')
    df3_string = df3.to_csv(index=False)

    with open(f'{args.output}/flowcell-summary.csv', 'w') as f:
        f.write(df1_string)
    with open(f'{args.output}/lane-summary.csv', 'w') as f:
        f.write(df2_string)
    with open(f'{args.output}/top-unknown-barcodes.csv', 'w') as f:
        f.write(df3_string)



from .. import api

description = """
Summarize a BED file.

This command will compute various summary statistics for a BED file. The
returned statistics include the total numbers of probes and covered base
pairs for each chromosome.

By default, covered base pairs are displayed in bp, but if you prefer you
can, for example, use '--bases 1000' to display in kb.
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        help='Summarize a BED file.',
    )
    parser.add_argument(
        'bed',
        help='BED file.'
    )
    parser.add_argument(
        '--bases',
        metavar='INT',
        type=int,
        default=1,
        help='Number to divide covered base pairs (default: 1).'
    )
    parser.add_argument(
        '--decimals',
        metavar='INT',
        type=int,
        default=0,
        help='Number of decimals (default: 0).'
    )

def main(args):
    bf = api.pybed.BedFrame.from_file(args.bed)
    chrom_dict = {}
    total = [0, 0]
    for i, r in bf.gr.df.iterrows():
        chrom = r['Chromosome']
        start = r['Start']
        end = r['End']
        bases = end - start
        if chrom not in chrom_dict:
            chrom_dict[chrom] = [0, 0]
        chrom_dict[chrom][0] += 1
        chrom_dict[chrom][1] += bases
        total[0] += 1
        total[1] += bases
    print('Chrom', 'Probes', 'Bases', sep='\t')
    for chrom in chrom_dict:
        results = chrom_dict[chrom]
        probes = results[0]
        bases = f'{results[1]/args.bases:.{args.decimals}f}'
        print(chrom, probes, bases, sep='\t')
    probes = total[0]
    bases = f'{total[1]/args.bases:.{args.decimals}f}'
    print('Total', probes, bases, sep='\t')



import sys

from .. import api

import pandas as pd

description = """
Filter a VCF file.
"""

epilog = f"""
[Example] Mark genotypes with 0/0 as missing:
  $ fuc {api.common._script_name()} in.vcf --expr 'GT == "0/0"' > out.vcf

[Example] Mark genotypes that are not 0/0 as missing:
  $ fuc {api.common._script_name()} in.vcf --expr 'GT != "0/0"' > out.vcf

[Example] Mark genotypes whose DP is less than 30 as missing:
  $ fuc {api.common._script_name()} in.vcf --expr 'DP < 30' > out.vcf

[Example] Same as above, but also mark ambiguous genotypes as missing:
  $ fuc {api.common._script_name()} in.vcf --expr 'DP < 30' --greedy > out.vcf

[Example] Build a complex query to select genotypes to be marked missing:
  $ fuc {api.common._script_name()} in.vcf --expr 'AD[1] < 10 or DP < 30' --opposite > out.vcf

[Example] Compute summary statistics and subset samples:
  $ fuc {api.common._script_name()} in.vcf \\
  --expr 'np.mean(AD) < 10' --greedy --samples sample.list > out.vcf

[Example] Drop duplicate rows:
  $ fuc {api.common._script_name()} in.vcf --drop_duplicates CHROM POS REF ALT > out.vcf

[Example] Filter out rows without genotypes:
  $ fuc {api.common._script_name()} in.vcf --filter_empty > out.vcf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Filter a VCF file.',
    )
    parser.add_argument(
        'vcf',
        help='VCF file (zipped or unzipped).'
    )
    parser.add_argument(
        '--expr',
        metavar='TEXT',
        help='Expression to evaluate.'
    )
    parser.add_argument(
        '--samples',
        metavar='PATH',
        help='File of sample names to apply the marking (one \n'
             'sample per line).'
    )
    parser.add_argument(
        '--drop_duplicates',
        metavar='TEXT',
        nargs='*',
        help='Only consider certain columns for identifying \n'
             'duplicates, by default use all of the columns.'
    )
    parser.add_argument(
        '--greedy',
        action='store_true',
        help='Use this flag to mark even ambiguous genotypes \n'
             'as missing.'
    )
    parser.add_argument(
        '--opposite',
        action='store_true',
        help='Use this flag to mark all genotypes that do not \n'
             'satisfy the query expression as missing and leave \n'
             'those that do intact.'
    )
    parser.add_argument(
        '--filter_empty',
        action='store_true',
        help='Use this flag to remove rows with no genotype \n'
             'calls at all.'
    )

def main(args):
    vf = api.pyvcf.VcfFrame.from_file(args.vcf)

    if args.expr is not None:
        if args.samples is None:
            samples = None
        else:
            samples = pd.read_table(args.samples, header=None)[0].to_list()

        vf = vf.markmiss(args.expr,
                         greedy=args.greedy,
                         opposite=args.opposite,
                         samples=samples)

    if args.drop_duplicates is not None:
        vf = vf.drop_duplicates(args.drop_duplicates)

    if args.filter_empty:
        vf = vf.filter_empty()

    sys.stdout.write(vf.to_string())
